/* Collection of procedures for estimating the Chow-Lin disaggregation. */
/* Also: a new concept of determing the autoregressive parameter. */
/* This collection is called by DisAggMain.ox. */
/* A safe version is 20070222ChowLin.ox. */


#include <oxstd.h>
#include <oxdraw.h>
#include <oxprob.h>
#import <modelbase>

static decl m_mY, m_mX, m_Sig, m_Rho, m_mXq, m_mYf, m_mZ, m_mZf, m_mzf;
static decl m_mM; // desaggregation coefficient
static decl mHf; // high frequency
static decl mLf; // low frequency

static decl s_mY, s_mX, s_cT;
static decl s_vAR, s_vMA, s_mM;
static decl s_dSigma, s_dVar;
static decl s_dPar, s_Zeig;
static decl s_Trend, s_CSd;

/* foreward declation of functions(), so that they can be used before it is typed */
/* NOTE: The order counts! */
MyChowLin(quelle, ziel, sofs, eofs, mTyn); // mTyn: Mit Trend (yes: 1, no:0)
MyMaxCL(const Rho, const adFunc, const avScore, const amHess); /*Maximises eq. 16 of Chow Lin wrt rho*/
MyDatIndex(const mM); /* Finds the index of the 1st observation */
MyResGraph(const mSamplStart, VarName, YName, ziel); /* Finds the index of the 1st observation */
TheChowLin(mY, mX, mXq, rho); /* Calculates the Chow-Lin-Estimate WITH estimation of rho */
MyChowLinCalculate(mY, mX, mXq); /* Calculates the Chow-Lin-Estimate without estimating (rho) */
MyCSeason(mT, mF); /* Calculates the centred seasonal dummies */

MyARMACL(const mY, const mX, const mAR, const mMA, const adLik, const fzeig);
MyFreeARMA(const mY, const mX, const mAR, const mMA, const adLik, const fzeig);
MyFreeAR(const mY, const mX, const mAR, const adLik, const fzeig);
MyFreeMA(const mY, const mX, const mMA, const adLik, const fzeig);
MyARMAstart(const mY, const mX, const s_Zeig);
SetCLParameters(const vP);
CLParametersM(const vP, const mM);
ArmaXLogLik(const vY, const vX, const pdLik);
LikelihoodCL(const vP, const pdLik, const pvSco, const pmHes); // original version
ArmaStderr(const vP);
MyARMAXgenH0(const beta, const rho, const mM);
MyARMAXgenH1(const beta, const rho);
GenAggARMAX(const beta, const rho, const mM);
GridSearch(const func, const avP, const adFunc, const cRange, const cPic); 
HillClimb(const func, const avP, const adFunc, const cRange, const cPic);
Stopit(); // stops program execution by moving into an endless loop, brak manually!
/* end of forward declaration */

MyChowLin(quelle, ziel, sofs, eofs, mTyn)
{
    decl b, covmat, XX, T, k, e, sigma2, V, SE, Tstat, mpar, R2, chowlin, j, mT;
	decl NAME, YVar, VarName, mSampl, sT, mCLRho, mCSd;
	decl vp, dlik, mRange, cMAa, cAr, cMa, cX, cChi, cChiMN, me, cXt, cR2, sCSD;
	decl jj = 0~mTyn[0];
	decl mj = jj[0];  // no. of Var?
	decl mTrend	= jj[1];
	decl mWSmpl = ones(2,2); // within sample dates e.g. the 2 in 1990q2

	s_CSd = mTyn[3];
	mHf = sofs[0][2];
	mLf = sofs[1][2];
	m_mM = mHf/mLf;

	// Load aggregated series
	mSampl = frequenz(sofs[1][], eofs[1][]); // low frequency data
	decl agseries = new Database();
	agseries.LoadXls(quelle[1][0]); // .xls has to have an unlabelled column (with the dates)

	// Create output database, frequency: mHf
	decl resultate = new Database();
	resultate->Create(mHf, mSampl[0], 1, mSampl[2], mHf); 

	// Find the available data range
	decl sYNAME = agseries->GetAllNames();
	decl mY = agseries.GetVar(sYNAME[0]);	 // sYNAME: aggregated series (includes missings)
	b = MyDatIndex(mY);
		/* 1 positions where subsample starts (alternatig with NANs) */
		/* 2 size of largest subsample starts */
		/* 3 positions where 1 yields this information */
	k = b[0];


	//effective sample based on availability of aggregated series :
	XX = {agseries.ObsYear(k[b[2][0]][0]), agseries.ObsYear(k[b[2][0]][0]+b[1][0]-1)};
	if (s_CSd==1){
	mWSmpl[0][1] = floor(mLf*(mSampl[0]+k[b[2][0]][0]/mLf - XX[0]))+1;
	}

	mSampl = frequenz(sofs[0][], eofs[0][]); // high frequency data

	chowlin = new Database();
	chowlin.LoadXls(quelle[0][0]);

	NAME = chowlin->GetAllNames();
	chowlin.Deterministic(TRUE);
	decl mChLinTest = new array[rows(NAME)]; 
	decl mErgebniss = zeros(rows(NAME), 13);
	j = 0;
	while (j<rows(NAME))
	{
	mj = j;
	VarName = NAME[mj]; /* reference series */
	
	print("\n");
	print("*************************************************************** \n");
	print("*********** Chow-Lin disaggregation for: ",VarName," ********** \n");
	print("*************************************************************** \n");

	decl Cmat, mX, mU, Rho, mTT, mV, m_mXqf, mXf, mUf, CmatF; //, mpar, mmpar, R2, tstats, coeffs;

	mX = chowlin.GetVar(VarName);	 // sYNAME: aggregated series (includes missings)
	b = MyDatIndex(mX[][0]);	// returns matrix with index of non NaNs
	k = b[0];
	e = k[b[2][0]][0]+b[1][0]-1;
	k = chowlin.ObsYear(k[b[2][0]])+1*(chowlin.ObsPeriod(k[b[2][0]])!=1);
	k = {k, (chowlin.ObsYear(e)-1*(chowlin.ObsPeriod(e)!=mHf)) };
	mWSmpl[1][1] = floor( (mHf*(mSampl[0]+b[0][b[2]][0]/mHf - k[0])))+1;


	mWSmpl[0][0] = maxc(XX[0]|k[0]); // common start year
	mWSmpl[1][0] = minc(XX[1]|k[1])-mTyn[1]; // common final year less forecast years

	mWSmpl[0][1] = maxc((mWSmpl[0][1])|mWSmpl[1][1]/m_mM);
	mWSmpl[1][1] = (mWSmpl[0][1]-1)*m_mM+1; 


	agseries.Select(Y_VAR, {sYNAME[0], 0, 0} );	 
	chowlin.Select(X_VAR, {VarName, 0, 0} );	 // NAMEk related series
	chowlin.Select(U_VAR,  {"Constant", 0, 0} );	 // Constant

	if (s_CSd==1){
	mCSd = MyCSeason( (mWSmpl[1][0]-mWSmpl[0][0]+m_mM)*mHf, mHf|mLf); /* Calculates the centred seasonal dummies */
	}

	
	// save X and Y for forecasting
	agseries.SetSample(mWSmpl[0][0],mWSmpl[0][1],mWSmpl[1][0]+mTyn[1],mWSmpl[0][1]); //	dbase->SetSample(1991,3,2003,3);
	m_mYf = agseries->GetGroup(Y_VAR);

//	chowlin.SetSample(mWSmpl[0][0],mWSmpl[1][1],mWSmpl[1][0]+mTyn[1], mWSmpl[1][1]-1); //	dbase->SetSample(1991,3,2003,3);
//	chowlin.SetSample(mWSmpl[0][0],mWSmpl[1][1],mWSmpl[1][0]+mTyn[1], mWSmpl[1][1]+m_mM*s_CSd-1); //	dbase->SetSample(1991,3,2003,3);
	chowlin.SetSample(mWSmpl[0][0],mWSmpl[1][1],mWSmpl[1][0]+mTyn[1], mWSmpl[1][1]+m_mM-1); //	dbase->SetSample(1991,3,2003,3);

	m_mXqf = chowlin->GetGroup(X_VAR);
	mUf = chowlin->GetGroup(U_VAR);
		if (s_CSd==1){
		mUf = mUf~mCSd[0:rows(mUf)-1][];
		}


	// chose common sample variables
	agseries.SetSample(mWSmpl[0][0],mWSmpl[0][1],mWSmpl[1][0],mWSmpl[0][1]); //	dbase->SetSample(1991,3,2003,3);
//	chowlin.SetSample(mWSmpl[0][0],mWSmpl[1][1],mWSmpl[1][0],mWSmpl[1][1]-1); //	dbase->SetSample(1991,3,2003,3);
//	chowlin.SetSample(mWSmpl[0][0],mWSmpl[1][1],mWSmpl[1][0],mWSmpl[1][1]+m_mM*s_CSd-1); //	dbase->SetSample(1991,3,2003,3);
	chowlin.SetSample(mWSmpl[0][0],mWSmpl[1][1],mWSmpl[1][0],mWSmpl[1][1]+m_mM-1); //	dbase->SetSample(1991,3,2003,3);

	mY = agseries->GetGroup(Y_VAR);	
	mX = chowlin->GetGroup(X_VAR);
	mU = chowlin->GetGroup(U_VAR);
		if (s_CSd==1){
		mU = mU~mCSd[0:rows(mU)-1][];
		}
	
	/* Aggregate X variables */
	mT = rows(mY);
	Cmat = 1/m_mM*unit(mT)**ones(1,m_mM); /* Aggregation matrix */
	CmatF = 1/m_mM*unit(mT+mLf*mTyn[1])**ones(1,m_mM); /* Aggregation matrix */

	if (mTrend == 1)
	{
	m_mXq = mX~mU[][]~(range(1, m_mM*mT, 1)');
	m_mXqf = m_mXqf~mUf[][]~(range(1, (m_mM*mT+mHf*mTyn[1]), 1)');
	mX = Cmat*(mX~mU[][])~(Cmat*(range(1, m_mM*mT, 1)'));
	}
	else
	{
	m_mXq = mX~mU;
	m_mXqf = m_mXqf~mUf;
	mX = Cmat*(mX~mU);
	}
	mXf = CmatF*m_mXqf;
	
	/*  Call ARMA for starting values */
	decl CMARMA;
	CMARMA = new ArfimaCL();		   // now an Arfima object
	CMARMA.SetPrint(mTyn[2]);
	//------------------------------------------------------- create the model
	CMARMA.DeSelect();										
	CMARMA.Append(mY, "Y", 0);                  // formulate model
    CMARMA.Append(mX[][0], "X", 0);                  // formulate model
	CMARMA.Append(mX[][1], "C", 0);                  
    CMARMA.Select(Y_VAR, {"Y", 0, 0} );                  // formulate model
    CMARMA.Select(X_VAR, {"X", 0, 0});                  
    CMARMA.Select(X_VAR, {"C", 0, 0});                  
	if (s_CSd==1)
	{
	CMARMA.Append(mX[][2:mLf], {"CSeason", "CSeason_1", "CSeason_2"}, 0);                  
	CMARMA.Select(X_VAR, {"CSeason", 0, 0, "CSeason_1", 0, 0, "CSeason_2", 0, 0});                  
	}
//	if (mHf==4)
//  CMARMA.Select(X_VAR, {"CSeason", 0, 0, "CSeason_1", 0, 0, "CSeason_2", 0, 0, "CSeason_3", 0, 0});                  
	if (mTrend == 1)
	{
    CMARMA.Append(mX[][columns(mX)-1], "T", 0);                  
    CMARMA.Select(X_VAR, {"T", 0, 0});                  
	}
//	CMARMA.Info();
	CMARMA.ARMA(1,0);
	CMARMA.FixD(0);// make sure co-integration
	CMARMA.SetMethod(M_MAXLIK);
	CMARMA.UseSampleMean();
	CMARMA.SetSelSample(-1, 1, -1, 1); //	dbase->SetSample( max. sample );
	CMARMA.Estimate();
	Rho = CMARMA.GetPar();
	Rho = fabs(Rho[1])^(1/m_mM)*(-1^(Rho[1]<0));	// 'disaggregated' Rho
	delete CMARMA;
/* END of preliminary ARMA(1,0) estimation */


//ARMAX() estimation
//Unrestricted ARMA(1,1)
	//Starting values
		cAr= -99;
		cMa=-99;
		cX=-99;
		cXt = -99;
		cR2 = -99;
		cChi =-99;  // 1 if rejected
		cChiMN=-99;  // marg. sign. level

	s_dPar = zeros(3,2+columns(mX)+mTyn[0]); // collects AR parameter at ``annual level''
	s_mM = m_mM; // alternative: monthly to annual
	s_mY = mY;
	s_mX = mX;
	s_cT = mT;
	s_Trend = mTyn[0];
	decl RhoUR = MyARMAstart(s_mY, s_mX, (mTyn[2]!=0)); // returns AR parameter at ``quarterly'' level
	RhoUR = RhoUR^m_mM;
	MaxControlEps(1e-5, 1e-3);
	MyFreeARMA(s_mY, s_mX, &vp, &cMAa, &dlik, (mTyn[2]!=0)); //dlik order: lik, AR, MA, parameters on regressors (X, const)
	MyARMACL(s_mY, s_mX, vp, cMAa, &dlik, 0); // evaluate at vP
	decl cfail = isnan(vp);
	if (cfail == 0) // everything fine
	{
	me = dlik[0];
	cAr = (fabs(dlik[1])^(1/m_mM))*((-1)^(1-(dlik[1]>0))); // rho
	cMa = dlik[2];
	cX  = dlik[3];
	decl cPrec = 1e-5;
//	mRange = {(-1*(( (-1)^m_mM)<0)+1e-5),(1-1e-5),30};
//	mRange = {(-1+1e-5),(1-1e-5),30};
	mRange = {(1e-5),(1-1e-5),30}; //The loglikelihood function is perfectly symmetric around zero!!
	GridSearch(LikelihoodCL, &vp, &dlik, mRange, (mTyn[2]!=0));
	decl mGSdlik = dlik*s_cT;
	//Local Optimisation
	HillClimb(LikelihoodCL, &vp, &dlik, (vp|cPrec), (mTyn[2]!=0));	 //last argument: starting value, increment, convergence criterion
	//Get parameters
	//Calculate benchmark size
	decl cA = CLParametersM(vp[0], m_mM); // map parameters to CL, assumes AR coefficient of quarterly level
	MyARMACL(s_mY, s_mX, s_vAR, s_vMA, &dlik, (mTyn[2]!=0));
	cAr = s_vAR; //mean
	cMa = s_vMA; //dlik[2];
	cX  = dlik[3];
	cXt = s_dPar[1][0];
	cR2 = s_dPar[2][0];
/* END ARMAX estimation */
	if (me>dlik[0])
	{
		cChi = (1-probchi(2*(me-dlik[0]),1)<0.05)*(1-(me==0));  // 1 if rejected
		cChiMN = (1-probchi(2*(me-dlik[0]),1));  // marg. sign. level
	print("\n Test successful!! \n");
	}		
	else
	{
	print("\n Test failed!! \n");
	cChi = 0;
	cChiMN = 1;
	}
//	if ((1-probchi(2*(me-dlik[0]),1)<cSN)*(1-(me==0)))??
	} //cfail

	
/*  Start constructing the ChowLin-Estimator*/
	decl r, mBeta, mBetaf, mA, mUhat, mZ, mZf;
	m_mY = mY;
	m_mX = mX;

/*  Estimating rho and beta with BFGS (!) */
	mBeta = TheChowLin(m_mY, m_mX, m_mXq, Rho );
	mZ = mBeta[1];
	m_mZ = mZ;
//	m_Rho = mBeta[2]; //  new rho
	mCLRho = mBeta[2]; //  new rho
	mBeta = mBeta[0];
	r = mBeta[1];     // R^2
	mBeta = mBeta[0];
	mZf = m_mXqf*mBeta[][0];
	

/* That kind of forecasting does not make sense! */
/* It does make sense though to re-estimate beta this way. */
	mA = MyChowLinCalculate(m_mYf, mXf, m_mXqf);
	mBetaf = mA[0];
	mBetaf = mBetaf[0];
	mBetaf = mBetaf[0][0];



	if (mTyn[1]>0) // incl. Forecast?
	{
	m_mZf = mZ[0:(rows(mZ)-1)][]|mZf[rows(mZ):(rows(mZf)-1)];
	Cmat = 1/m_mM*unit(mT+mLf*mTyn[1])**ones(1,m_mM); /* Aggregation matrix */
	m_mzf = Cmat*mZf;
	Cmat = 1/m_mM*unit(mT)**ones(1,m_mM); /* Aggregation matrix */
	mA = Cmat*mZ;
	m_mzf = mA[0:(rows(mA)-1)][]|m_mzf[rows(mA):(rows(m_mzf)-1)];
	mA = m_mzf[rows(mA):(rows(m_mzf)-1)]-m_mYf[rows(mA):(rows(m_mzf)-1)];
	mA = (mA'mA)/(mTyn[1]*m_mY[(mT-1)]);
	mBetaf = fabs(mBeta[0][0]-mBetaf)/mBeta[0][0]; // percentage change of Beta w.r.t. to sample
	}
	else
	{
	m_mzf = m_mY;
	mA = -99;
	mBetaf =-99;
	}


	if (mTyn[2]!=0)
	{
	m_Rho = cAr; // set global Rho to MaxLik estimator!
	MyResGraph(mWSmpl[0][0], VarName, sYNAME[0], ziel[0]);
	}

	// store results
	mChLinTest[j] = {VarName, mBeta[][0]'~mBeta[][1]'~r};
	mErgebniss[j][0:12] = mBeta[0][0]'~fabs(mBeta[0][2])'~Rho^(m_mM)~mCLRho~r~cAr~cX~cXt~cR2~cChiMN~sqrt(variance(mY))/(fabs(meanc(mY)))~sqrt(variance(mZ))/(fabs(meanc(mZ)))~mA~mBetaf~rows(m_mY);
	resultate.SetSelSample(mSampl[0], 1, mSampl[2], 4);
	resultate->Append(mZ, {sYNAME[0]~VarName}, resultate.GetIndex(mSampl[0], mSampl[1]));

		agseries.DeSelect();
		chowlin.DeSelect();

		++j;		  // End of loop for all related series
	  }

	decl sTrendPresence;
	if (mTrend == 1)
	{
	sTrendPresence = "_T_";
	resultate.SaveXls(ziel[0]~"_T.xls");
	sT = "";
	}
	else
	{
	sTrendPresence = "_noT_";
	resultate.SaveXls(ziel[0]~".xls");
	sT = "out";
	}
	if (mTyn[3] == 1)  // Centred seasonal dummies?
	{
	sTrendPresence = sTrendPresence~"cSD";
	sCSD = "";
	}
	else
	{
	sTrendPresence = sTrendPresence~"NOcSD";
	sCSD = "out";
	}
	sTrendPresence = sTrendPresence~sprint(mLf)~"to"~sprint(mHf);

			format(600);
			sprintbuffer(20480);
			decl Names = mChLinTest[0];
			mChLinTest= mChLinTest[1];
			
            print("\nChow-Lin-disaggregation for the Variable ", sYNAME[0], " for ", XX[0] , " to ", XX[1], "\n",
            "Trend: with"~sT~"\n",
            "Centred sesonal dummies: with"~sCSD~"\n",
            "Disaggregate level: ", mHf, " (periods per year). \n",
            "Aggregate level: ", mLf, " (Periods per year). \n",
            "Number of out-of-sample forecasts: ", mTyn[1], ". \n",
            "%r", NAME, //CiAns
            "%cf", {"%d2.0g",  "%d2.2g",    "%d2.2g",    "%d2.2g",     "%d2.2g",    "%d2.2g",    "%d2.2g",     "%d2.2g",    "%d2.2g",    "%d2.2g",    "%d2.2g",  "%d2.2g",   "%d2.2g"},
            "%c",  {"Beta(X)" , "t-val  ", "rho(AR(1))", "  rho(CL)  " , "R^2(CL)  ", "rho(CM)  ", "Beta(X)CM", "t-val  " , "R^2(CM)  ", " chi^2(1) ", "Std(Ylf)  ", "Std(Yhf)  ",  "E(F)^2", "Var(b)",   " Obs   "},
            "%cf", {" %8.4g ", "  %8.4g  ", " %8.4g ",   " %10.4g "  , "  %8.4g  ", " %8.4g ",   " %10.4g "   , "  %8.4g  ", "  %8.4g  ",  "  %8.4g ",  "  %8.4g ",  "  %8.4g ",  " %8.4g ", " %8.4g ",  " %5.4g "  },
            mErgebniss,                                                                          
            "\nThese results (without header) are to be found at:\n", ziel[0]~sTrendPresence~".dat",   
            "\nThis file: ", ziel[0]~sTrendPresence~".txt");           // ResCoeff %#d                    

	decl ergebnis = sprint(
			"%r", NAME, //CiAns
            "%cf", {"%d2.0g",  "%d2.2g",    "%d2.2g",    "%d2.2g",     "%d2.2g",    "%d2.2g",    "%d2.2g",     "%d2.2g",    "%d2.2g",    "%d2.2g",    "%d2.2g",  "%d2.2g",   "%d2.2g"},
            "%c",  {"Beta(X)" , "t-val  ", "rho(AR(1))", "  rho(CL)  " , "R^2(CL)  ", "rho(CM)  ", "Beta(X)CM", "t-val  " , "R^2(CM)  ", " chi^2(1) ", "Std(Ylf)  ", "Std(Yhf)  ",  "E(F)^2", "Var(b)",   " Obs   "},
            "%cf", {" %8.4g ", "  %8.4g  ", " %8.4g ",   " %10.4g "  , "  %8.4g  ", " %8.4g ",   " %10.4g "   , "  %8.4g  ", "  %8.4g  ",  "  %8.4g ",  "  %8.4g ",  "  %8.4g ",  " %8.4g ", " %8.4g ",  " %5.4g "  },
			mErgebniss);
	decl file = fopen((ziel[0]~sTrendPresence~".txt"), "w");
	fprint(file, ergebnis);
	fclose(file);		

	ergebnis = sprint("\nChow-Lin-disaggregation for the Variable ", sYNAME[0], " for ", XX[0] , " to ", XX[1], "\n",
            "Trend: with"~sT~"\n",
            "Centred sesonal dummies: with"~sCSD~"\n",
            "Disaggregate level: ", mHf, " (periods per year). \n",
            "Aggregate level: ", mLf, " (Periods per year). \n",
	        "Number of out-of-sample forecasts: ", mTyn[1], ". \n",
			"%r", NAME, //CiAns
            "%cf", {"%d2.0g",  "%d2.2g",    "%d2.2g",    "%d2.2g",     "%d2.2g",    "%d2.2g",    "%d2.2g",     "%d2.2g",    "%d2.2g",    "%d2.2g",    "%d2.2g",  "%d2.2g",   "%d2.2g"},
            "%c",  {"Beta(X)" , "t-val  ", "rho(AR(1))", "  rho(CL)  " , "R^2(CL)  ", "rho(CM)  ", "Beta(X)CM", "t-val  " , "R^2(CM)  ", " chi^2(1) ", "Std(Ylf)  ", "Std(Yhf)  ",  "E(F)^2", "Var(b)",   " Obs   "},
            "%cf", {" %8.4g ", "  %8.4g  ", " %8.4g ",   " %10.4g "  , "  %8.4g  ", " %8.4g ",   " %10.4g "   , "  %8.4g  ", "  %8.4g  ",  "  %8.4g ",  "  %8.4g ",  "  %8.4g ",  " %8.4g ", " %8.4g ",  " %5.4g "  },
			mErgebniss,
			"\nThese results (without header) are to be found at:\n", ziel[0]~sTrendPresence~".dat",
			"\nThis file: ",  ziel[0]~sTrendPresence~".txt"); 		  // ResCoeff %#d
 
	file = fopen((ziel[0]~sTrendPresence~".txt"), "a");
	fprint(file, ergebnis);
	fclose(file);		


	return 1;
	
	delete chowlin;
}  //End of MyChowLin

// Function to be maximised (by BFGS in TheChowLin() )
//Bprobit::fProbit(const vP, const adFunc, const avScore, const amHess)
MyMaxCL(const rho, const adFunc, const avScore, const amHess) /*Maximises eq. 16 of Chow Lin wrt rho*/
{
	decl b, mT, mV, mSca, e, r, Cmat, mX, mY, mXq, mZ;
	mSca = (fabs(maxc(m_mY))+fabs(minc(m_mY)))/2;
	mT = rows(m_mY);
	mY = m_mY/mSca;
	mX = m_mX/mSca;
	Cmat = 1/m_mM*unit(mT)**ones(1,m_mM); /* Aggregation matrix */	
	b = range(1, m_mM*mT, 1); 
	r = ones(1, m_mM*mT)*rho;
	mV = (r.^(-b))'(r.^b);
	e = mV';
	mV = setlower(mV,e);
	mV = (unit(mT)-mX*invert(mX'invert(Cmat*mV*Cmat')*mX)*mX'*invert(Cmat*mV*Cmat'))*mY; // is U 
	mZ = lag0(mV,1);
	mZ = invert(mZ'mZ)*mZ'*mV; // calculate update of rho and
	mV = mZ - rho;
	adFunc[0] = -1*(mV'mV)*10; // min sum squared, factor 10 simplifies numerical evaluation
	return 1;
}	

MyDatIndex(const mM) /* Finds the index of the 1st and last observations */
// Input: mxs matrix
// Output: 1. nxs matrix with row indexes
//						positive entries represent start of a sequence of observations that are not NAN
//						and negative entries represent start of a sequence of NANs
//			2. 1xs matrix with max spells
//			3. 1xs matrix with index of start of longest spell in output 1
{
	decl mR, mC, mRi, mCi, mA, mi, mii, mMM, mj, ma, mD, mDD;
	mR = <0, 0>;
	mj = 0;
	mMM = -1*(isdotnan(mM)+(isdotnan(mM)-1));
	mC = zeros(1, columns(mM));
	mD = mC;
	for (mii = 0; mii<columns(mM); mii++) // columns search
	{
	mCi = 0;
		if (sumc(isdotnan(mM[][mii]).==0)<rows(mM[][mii]))
		{
		ma=1;
		for (mi = 0; mi<(rows(mM)-1); mi++)	  // row search
		{
// Regime change?: mMM[mi][mii]*(mMM[mi][mii]==mMM[fabs(mC[mCi][mii])][mii]);
		mC[mCi][mii] = mC[mCi][mii] + mMM[mi][mii]*(mMM[mi][mii]==mMM[fabs(mC[mCi][mii])][mii]);
		if (mMM[mi][mii] != mMM[fabs(mC[mCi][mii])][mii])
			{
			mC = mC|(zeros(1, columns(mMM)));
				mC[mCi+1][mii] = -1*mC[mCi][mii];
			mCi = mCi + 1; // counts the number of regime changes
			}
			else if((mi==0)&&(mC[mCi][mii]==1))
			{
			mD[0][mii]=-1;
			}
		 }

		 }
		 else
		 {
		 ma = 0;
		 }
	}

	if (ma==1)
	{
	mC = mD|mC;
	mC = selectifr(-1*mC, mC.!=0);	
	mC[0][] = mC[0][]-(mD.==-1);
	mRi = selectifr(mC, mC.!=0);
	mj = fabs(mRi)-fabs(lag0(mRi,1));
	return {mC, maxc(mj.*(mRi.<0)), (maxcindex(mj.*(mRi.<0))-1)+1*(mD.==-1) };			  //
	}
	else
	{
	return { (0|(-1*rows(mM))), rows(mM), 0 };
	}

}

MyResGraph(const mSamplStart, VarName, YName, ziel)
{
	decl mT, Cmat, mA;
	mT = rows(m_mY);

	DrawTitle(0, "Disaggregation of "~YName~" using "~VarName~" as related series (aggregated data, rho="~sprint(m_Rho)~")");
	DrawLegend(0, 20, 140, 0);
	DrawTMatrix(0, (m_mzf)', {"Chow-Lin estimates"}, mSamplStart, 1, mLf, 0, 2); // }, year, period, frequency, symbol, index
	DrawTMatrix(0, (m_mYf)', {"original series"}, mSamplStart, 1, mLf, 2, 4); //mSamplStart-1, 1, 1, 2, 4);

//	DrawAxisAuto(1, 0, minc(m_mZ), maxc(m_mZf), minc(m_mZ), 1, mLf, mHf, TRUE, ANCHOR_USER);
	DrawTitle(1, "Disaggregated data including forecasts (if applicable)");
	DrawTMatrix(1, (m_mZf)', {"forecasts"}, mSamplStart, 1, mHf, 0, 2); // mSamplStart-1, 1, 4, 0, 2);
	DrawTMatrix(1, (m_mZ)', {"Chow-Lin estimates"}, mSamplStart, 1, mHf, 2, 4); //mSamplStart-1, 1, 4, 2, 4);

	DrawTitle(2, "Forecast comparison (if applicable)");
	DrawTMatrix(2, (m_mzf[rows(m_mY)-4*mLf:])', {"Chow-Lin forecasts"}, mSamplStart+(rows(m_mY)-4*mLf)/mLf, 1, mLf, 0, 2);
	DrawTMatrix(2, (m_mYf[rows(m_mY)-4*mLf:])', {"original series"}, mSamplStart+(rows(m_mY)-4*mLf)/mLf, 1, mLf, 2, 4);

	ShowDrawWindow();
	SaveDrawWindow(ziel~"Graf"~VarName~".eps");

}

// Estimation of rho and beta
TheChowLin(mY, mX, mXq, rho)
{
	decl r, mBeta, mA, mUhat, mZ, Rho, mT, mV, e, Cmat;
	m_mY = mY;
	m_mX = mX;
	m_mXq = m_mXq;
	Rho = rho;
	MaxControl(500, 0, TRUE); // sets iteration infos: max Iterations, printe every ..., use compact format if true
	MaxControlEps(0.0001, 0.001);
	m_Rho = MaxBFGS(MyMaxCL, &Rho, &m_Sig, 0, TRUE);   /* uses numerical 1st derivatives if TRUE */
	mT = rows(m_mXq);
	mA = range(1, mT, 1);
	r = ones(1, mT)*Rho;
	mV = (r.^(-mA))'(r.^mA);
	e = mV';
	mV = setlower(mV,e);
	mZ = mV;
	Cmat = 1/m_mM*unit(mT/m_mM)**ones(1,m_mM); /* Aggregation matrix */
	mV = Cmat*mV*Cmat';
	mZ = m_mXq*invert(m_mX'*invert(mV)*m_mX)*m_mX'*invert(mV)*m_mY + ((mZ*Cmat')*invert(mV))*(m_mY - m_mX*invert(m_mX'*invert(mV)*m_mX)*m_mX'*invert(mV)*m_mY);
	mBeta = invert(m_mX'*invert(mV)*m_mX)*m_mX'*invert(mV)*m_mY;
    mUhat = ((unit(mT/m_mM)-m_mX*invert(m_mX'invert(mV)*m_mX)*m_mX'*invert(mV))*m_mY)'*((unit(mT/m_mM)-m_mX*invert(m_mX'invert(mV)*m_mX)*m_mX'*invert(mV))*m_mY)*m_mM/(mT-columns(m_mX));
	mBeta = {(mBeta~sqrt(diagonal(mUhat*invert(m_mX'*invert(mV)*m_mX)))'~(mBeta./(sqrt(diagonal(mUhat*invert(m_mX'*invert(mV)*m_mX))')))), 1-mUhat/variance(m_mY)};
return {mBeta, mZ, Rho};
//Output: mBeta: beta coefficient on explanatory, s.d. of estimate, t-value

}

// Calculation of the disaggregated time series
MyChowLinCalculate(mY, mX, mXq)
{
	decl r, mBeta, mA, mUhat, mZ, mT, mV, e, Cmat;
	mT = rows(mXq);
	mA = range(1, mT, 1);
	r = ones(1, mT)*m_Rho;
	mV = (r.^(-mA))'(r.^mA);
	e = mV';
	mV = setlower(mV,e);
	mZ = mV;
	Cmat = 1/m_mM*unit(mT/m_mM)**ones(1,m_mM); /* Aggregation matrix */
	mV = Cmat*mV*Cmat';
	mZ = mXq*invert(mX'*invert(mV)*mX)*mX'*invert(mV)*mY + ((mZ*Cmat')*invert(mV))*(mY - mX*invert(mX'*invert(mV)*mX)*mX'*invert(mV)*mY);
	/*Note: the trend coefficient is 1/T of that in the ARMA estimation (T: no. obs. of aggregated series)   */
	mBeta = invert(mX'*invert(mV)*mX)*mX'*invert(mV)*mY;
    mUhat = ((unit(mT/m_mM)-mX*invert(mX'invert(mV)*mX)*mX'*invert(mV))*mY)'*((unit(mT/m_mM)-mX*invert(mX'invert(mV)*mX)*mX'*invert(mV))*mY)*m_mM/(mT-columns(mX));
//    mUhat = ((unit(mT/4)-mX*invert(mX'invert(mV)*mX)*mX'*invert(mV))*mY)'*((unit(mT/4)-mX*invert(mX'invert(mV)*mX)*mX'*invert(mV))*mY)*4/(mT-columns(mX)); 
	mBeta = {(mBeta~sqrt(diagonal(mUhat*invert(mX'*invert(mV)*mX)))'~(mBeta./(sqrt(diagonal(mUhat*invert(mX'*invert(mV)*mX))')))), 1-mUhat/variance(mY)};
return {mBeta, mZ};
}


MyARMACL(const mY, const mX, const mAR, const mMA, const adLik, const fzeig)
{
    decl b, c, d;
	decl CMARMA;
	CMARMA = new ArfimaCL();		   // dbase is now a PcFiml object!!
	CMARMA.SetPrint(fzeig);
	//------------------------------------------------------- create the model
	CMARMA.Create(1, 1, 1, rows(mY), 1);
	CMARMA.DeSelect();										// formulate model
	CMARMA.Append(mY , "YVAR", 0); //(mSampl[4]-1)/6-4, 168
	CMARMA.Select(Y_VAR, {"YVAR", 0, 0} );	 // endog. series in 1st difference,
	CMARMA.Append(mX[][0] , "XVAR", 0); //(mSampl[4]-1)/6-4, 168
	CMARMA.Select(X_VAR, {"XVAR", 0, 0});	 // endog. series in 1st difference,
	CMARMA.Deterministic(FALSE);
	CMARMA.Select(X_VAR, {"Constant", 0, 0});	 // constant term added,
	if (s_CSd==1)
	{
	CMARMA.Append(mX[][2:mLf], {"CSeason", "CSeason_1", "CSeason_2"}, 0);                  
	CMARMA.Select(X_VAR, {"CSeason", 0, 0, "CSeason_1", 0, 0, "CSeason_2", 0, 0});                  
	}
	if (s_Trend ==1)
	{
//	c = range(1,rows(mX),1);
//	CMARMA.Append(c', "Trend", 0);
	CMARMA.Select(X_VAR, {"Trend", 0, 0});	 // constant term added,
	}
    CMARMA.Arma(1,1);
	CMARMA.FixD(0);                  			  // make sure co-integration
	CMARMA.FixAR(1, mAR);
    CMARMA.FixMA(1, mMA);
	CMARMA.SetMethod(M_MAXLIK);                  			  // choose maximisation method
	CMARMA.Estimate();
	b = CMARMA.GetPar();
	c = CMARMA.GetSigma2();
	adLik[0] = (CMARMA.GetLogLik())~b[1:]';	 // /s_cT
	s_dPar[0][] = b[1:]'~c;
	d = (mY-meanc(mY))'(mY-meanc(mY))/rows(mY);
	d = c/d; 
	c = (rows(mY)-1)/(rows(mY)-columns(mX)-s_Trend-1);
	d = 1-d*c; // korr. Pseudo R^2
	s_dPar[2][0] = d;
	c = CMARMA.GetStdErr();
	c = (b[3:]) ./ c[3:rows(c)-1];
	s_dPar[1][] = c';
	delete CMARMA;
	return <>; //must not return values	if to be used in MaxBFGS!!
}

MyFreeARMA(const mY, const mX, const mAR, const mMA, const adLik, const fzeig)
{
    decl b, c;
	decl CMARMA;
	CMARMA = new ArfimaCL();		   // dbase is now a PcFiml object!!
	CMARMA.SetPrint(fzeig);
//	CMARMA.SetPrint(1);
   //------------------------------------------------------- create the model
	CMARMA.Create(1, 1, 1, rows(mY), 1);
	CMARMA.DeSelect();										// formulate model
	CMARMA.Append(mY , "YVAR", 0); 
	CMARMA.Select(Y_VAR, {"YVAR", 0, 0} );	 
	CMARMA.Append(mX[][0] , "XVAR", 0); //(mSampl[4]-1)/6-4, 168
	CMARMA.Select(X_VAR, {"XVAR", 0, 0});	 // endog. series in 1st difference,
	CMARMA.Deterministic(FALSE);
	CMARMA.Select(X_VAR, {"Constant", 0, 0});	 // constant term added,
	if (s_CSd==1)
	{
	CMARMA.Append(mX[][2:mLf], {"CSeason", "CSeason_1", "CSeason_2"}, 0);                  
	CMARMA.Select(X_VAR, {"CSeason", 0, 0, "CSeason_1", 0, 0, "CSeason_2", 0, 0});                  
	}
	if (s_Trend ==1)
	{
//	c = range(1,rows(mX),1);
//	CMARMA.Append(c', "Trend", 0);
	CMARMA.Select(X_VAR, {"Trend", 0, 0});	 // constant term added,
	}
    CMARMA.ARMA(1,1);
	CMARMA.FixD(0);                  			  // make sure co-integration
	CMARMA.SetMethod(M_MAXLIK);                  			  // choose maximisation method, M_MAXMPLIK M_MAXLIK(NLS works)
	CMARMA.Estimate();
	b = CMARMA.GetFreePar(); //order: (d), AR, MA, parameters on regressors
	mAR[0] = b[0];
	mMA[0] = b[1];
	adLik[0] = (CMARMA.GetLogLik()/s_cT)~b';
	delete CMARMA;	
	return <>; //must not return values	if to be used in MaxBFGS!!
}

MyFreeAR(const mY, const mX, const mAR, const adLik, const fzeig)
{
    decl b, c;
	decl CMARMA;
	CMARMA = new ArfimaCL();		   // dbase is now a PcFiml object!!
	CMARMA.SetPrint(fzeig);
    //------------------------------------------------------- create the model
	CMARMA.Create(1, 1, 1, rows(mY), 1);
	CMARMA.DeSelect();										// formulate model
	CMARMA.Append(mY , "YVAR", 0); 
	CMARMA.Select(Y_VAR, {"YVAR", 0, 0} );	 
	CMARMA.Append(mX[][0] , "XVAR", 0); //(mSampl[4]-1)/6-4, 168
	CMARMA.Select(X_VAR, {"XVAR", 0, 0});	 // endog. series in 1st difference,
	CMARMA.Deterministic(FALSE);
	CMARMA.Select(X_VAR, {"Constant", 0, 0});	 // constant term added,
	if (s_CSd==1)
	{
	CMARMA.Append(mX[][2:mLf], {"CSeason", "CSeason_1", "CSeason_2"}, 0);                  
	CMARMA.Select(X_VAR, {"CSeason", 0, 0, "CSeason_1", 0, 0, "CSeason_2", 0, 0});                  
	}
	if (s_Trend ==1)
	{
	c = range(1,rows(mX),1);
	CMARMA.Append(c', "Trend", 0);
	CMARMA.Select(X_VAR, {"Trend", 0, 0});	 // constant term added,
	}
    CMARMA.ARMA(1,0);
	CMARMA.FixD(0);                  			  // make sure co-integration
	CMARMA.SetMethod(M_MAXLIK);                  			  // choose maximisation method, M_MAXMPLIK M_MAXLIK(NLS works)
	CMARMA.Estimate();
	b = CMARMA.GetFreePar(); //order: (d), AR, MA, parameters on regressors
	mAR[0] = b[0];
	adLik[0] = (CMARMA.GetLogLik()/s_cT)~b';
	delete CMARMA;	
	return <>; //must not return values	if to be used in MaxBFGS!!
}

MyFreeMA(const mY, const mX, const mMA, const adLik, const fzeig)
{
    decl b, c;
	decl CMARMA;
	CMARMA = new ArfimaCL();		   // dbase is now a PcFiml object!!
	CMARMA.SetPrint(fzeig);
    //------------------------------------------------------- create the model
	CMARMA.Create(1, 1, 1, rows(mY), 1);
	CMARMA.DeSelect();										// formulate model
	CMARMA.Append(mY , "YVAR", 0); 
	CMARMA.Select(Y_VAR, {"YVAR", 0, 0} );	 
	CMARMA.Append(mX[][0] , "XVAR", 0); //(mSampl[4]-1)/6-4, 168
	CMARMA.Select(X_VAR, {"XVAR", 0, 0});	 // endog. series in 1st difference,
	CMARMA.Deterministic(FALSE);
	CMARMA.Select(X_VAR, {"Constant", 0, 0});	 // constant term added,
	if (s_CSd==1)
	{
	CMARMA.Append(mX[][2:mLf], {"CSeason", "CSeason_1", "CSeason_2"}, 0);                  
	CMARMA.Select(X_VAR, {"CSeason", 0, 0, "CSeason_1", 0, 0, "CSeason_2", 0, 0});                  
	}
	if (s_Trend ==1)
	{
	c = range(1,rows(mX),1);
	CMARMA.Append(c', "Trend", 0);
	CMARMA.Select(X_VAR, {"Trend", 0, 0});	 // constant term added,
	}
    CMARMA.ARMA(0,1);
	CMARMA.FixD(0);                  			  // make sure co-integration
	CMARMA.SetMethod(M_MAXLIK);                  			  // choose maximisation method, M_MAXMPLIK M_MAXLIK(NLS works)
	CMARMA.Estimate();
	b = CMARMA.GetFreePar(); //order: (d), AR, MA, parameters on regressors
	mMA[0] = b[0];
	adLik[0] = (CMARMA.GetLogLik()/s_cT)~b';
	delete CMARMA;	
	return <>; //must not return values	if to be used in MaxBFGS!!
}


MyARMAstart(const mY, const mX, const s_Zeig)
{
	//generates starting values for BFGS by AR(1) model
    decl mpar, c;
	decl CMARMA;
	CMARMA = new ArfimaCL();		   // CMARMA is now a PcFiml object!!
	CMARMA.SetPrint(s_Zeig);
    //------------------------------------------------------- create the model
	CMARMA.Create(1, 1, 1, rows(mY), 1);
	CMARMA.DeSelect();										// formulate model
	CMARMA.Append(mY , "YVAR", 0); 
	CMARMA.Select(Y_VAR, {"YVAR", 0, 0} );	 
	CMARMA.Append(mX[][0] , "XVAR", 0); //(mSampl[4]-1)/6-4, 168
	CMARMA.Select(X_VAR, {"XVAR", 0, 0});	 // endog. series in 1st difference,
	CMARMA.Deterministic(FALSE);
	CMARMA.Select(X_VAR, {"Constant", 0, 0});	 // constant term added,
	if (s_CSd==1)
	{
	CMARMA.Append(mX[][2:mLf], {"CSeason", "CSeason_1", "CSeason_2"}, 0);                  
	CMARMA.Select(X_VAR, {"CSeason", 0, 0, "CSeason_1", 0, 0, "CSeason_2", 0, 0});                  
	}
	if (s_Trend ==1)
	{
//	c = range(1,rows(mX),1);
//	CMARMA.Append(c', "Trend", 0);
	CMARMA.Select(X_VAR, {"Trend", 0, 0});	 // constant term added,
	}
    CMARMA.ARMA(1,0);
	CMARMA.FixD(0);                  			  // make sure co-integration
	CMARMA.Estimate();
	mpar = 	CMARMA.GetFreePar(); //order: (d), AR, MA, parameters on regressors
	mpar = fabs(mpar[0])^(1/s_mM)*(-1^(mpar[1]<0));	// 'disaggregated' Rho
	delete CMARMA;
	return mpar; 
}

SetCLParameters(const vP)
{
	// map to Aggregation constrained
	s_vAR = vP[0]^4;
	s_vMA = ((4+6*vP[0]+8*vP[0]^2 +8*vP[0]^3 +8*vP[0]^4 +6*vP[0]^5 +4*vP[0]^6)/(2*(   vP[0]+2*vP[0]^2+4*vP[0]^3 +2*vP[0]^4+vP[0]^5))
	 - sqrt( (4+6*vP[0]+8*vP[0]^2 +8*vP[0]^3 +8*vP[0]^4 +6*vP[0]^5 +4*vP[0]^6 )^2/(4*(vP[0]+2*vP[0]^2+4*vP[0]^3 +2*vP[0]^4+vP[0]^5)^2 )-1 ) );
}

CLParametersM(const vP, const mM)
{
	// map to Aggregation constrained
	decl mH, mh1, mh2, mpsi, mS0, mS1, mm, mmh;
	mmh = vP[0];
	if (mmh/=0)
	{
	mh1 = upper(ones(mM,mM));
	mh1 = mh1[][0:mM-2];
	mh2 = lower(ones(mM,mM));
	mh2 = mh2[][1:mM-1];
	mH = mh1~ones(mM,1)~mh2;
	s_vAR = vP[0]^mM;
	mpsi = (vP[0]*ones(mM-1,1)).^range(mM-1, 0);
	mpsi = mpsi[0][]';
	mS0 = mpsi'*mH*mH'*mpsi;
	// adjust variance of u^*
	mS1 = mpsi'*mh1*mh2'*mpsi;
	mh1 = mS0/(2*mS1)-sqrt(mS0^2/(4*mS1^2)-1);
	mh2 = mS0/(2*mS1)+sqrt(mS0^2/(4*mS1^2)-1);
	mm = mh1|mh2;
	mh1 = mincindex(fabs(mm));
	s_vMA = mm[mh1];
	mS0 = mS0/(1+s_vMA^2);
	}
	else
	{
	mS0 = 1;
	s_vMA = 0;
	}
	return mS0;
}

ArmaXLogLik(const vY, const vX, const pdLik)
{ // ARMAX estimation
	decl ret_val;
	ret_val = MyARMACL(vY, vX, s_vAR, s_vMA, pdLik, s_Zeig); // evaluate at vP
	return ret_val; // 1 indicates success, 0 failure
}

LikelihoodCL(const vP, const pdLik, const pvSco, const pmHes) // original version
{ // arguments dictated by MaxBFGS()
	decl ret_val, a;
//	SetCLParameters(vP[0]^(1/4)); // map vP to CL model
	a = fabs(vP[0])^(1/s_mM)*(-1^(vP[0]<0));	// 'disaggregated' Rho
	a = CLParametersM(a, s_mM); // map vP to CL model
//	a = CLParametersM(vP[0]^(1/s_mM), s_mM); // map vP to CL model
	ret_val = ArmaXLogLik(s_mY, s_mX, pdLik); // evaluate at vP
	pdLik[0] /= s_cT; // log-likelihood scaled by sample size
	return ret_val; // 1 indicates success, 0 failure
}

ArmaStderr(const vP)
{
	decl covar, invcov, var = s_vAR, vma = s_vMA, dsig = s_dSigma, result;
//	result = Num2Derivative(Likelihood, vP, &covar);
	result = Num2Derivative(LikelihoodCL, vP, &covar);
	s_vAR = var, s_vMA = vma, s_dSigma = dsig; // reset after Num2Der
	if (!result)
		{ print("Covar() failed in numerical second derivatives\n");
	return zeros(vP);
	}
	invcov = invertgen(-covar, 30);
	return sqrt(diagonal(invcov) / s_cT)';
}

MyARMAXgenH0(const beta, const rho, const mM)
{
	decl me, cX, cRid, cVar;
	cRid = 1000;
	s_cT += cRid;
	me = rann(s_cT,2); //generates random normal numbers
	me[][1] *= 1; // pretend exogenous variable having been aggregated from Q-A
	cVar = CLParametersM(rho[0], mM); // map parameters to CL, assumes AR coefficient of quarterly level
//	SetCLParameters(rho); // map parameters to CL, assumes AR coefficient of quarterly level
	cX = zeros(s_cT,1);
	s_mY = armagen(cX, sqrt(cVar)*rho[1]*me[][0], (s_vAR~s_vMA), 1, 1)+ones(s_cT,1)*beta[0]+beta[1]*me[][1]; // generates fake data
	s_mY = s_mY[cRid:][];
	s_cT -= cRid;
	s_mX = me[cRid:][1];
	return <>;
}

MyARMAXgenH1(const beta, const rho)
{
	decl me, cX, cRid;
	cRid = 1000;
	s_cT += cRid;
	me = rann(s_cT,2); //generates random normal numbers
	me[][1] *= 1;
	cX = zeros(s_cT,1);
	s_mY = armagen(cX, rho[2]*me[][0], (rho[0]~rho[1]), 1, 1)+ones(s_cT,1)*beta[0]+beta[1]*me[][1]; // generates fake data
	s_cT -= cRid;
	s_mX = me[][1];
	return <>;
}

GenAggARMAX(const beta, const rho, const mM)
{
	decl me, cX, cRid, cSig;
	cRid = 1000;
	s_cT += cRid;
	me = rann(s_cT,2); //generates random normal numbers
	me[][1] *= 1;
	cSig = CLParametersM(rho[0], mM); // map parameters to CL, assumes AR coefficient of quarterly level
	cX = zeros(s_cT,1);
	s_mY = armagen(cX, sqrt(cSig)*rho[1]*me[][0], (s_vAR~s_vMA), 1, 1)+ones(s_cT,1)*beta[0]+beta[1]*me[][1]; // generates fake data
	s_mY = s_mY[cRid:][];
	s_cT -= cRid;
	s_mX = me[cRid:][1];
	return <>;
}


GridSearch(const func, const avP, const adFunc, const cRange, const cPic) 
{ // arguments dictated by MaxBFGS()
	// function, parameter to be changed, criterion, parameter range
	decl i, mA, mB, mC, mAux;
	mA = (0|0);
	for (i = cRange[0]; i <= cRange[1]; i += (cRange[1]-cRange[0])/cRange[2])
	{
	func(i^s_mM, &mB, &mAux, 1);
	mB = mB[0];
	mA = mA~(i|mB);
	}
	mA = mA[][1:]; //discard double zero
	mC = maxcindex(mA[1][]');
	avP[0] = mA[0][mC];	  //rho (quarterly level)
	adFunc[0] =  mA[1][mC];  //loglik
	mB = mincindex(mA[1][]');
	CloseDrawWindow();
	DrawTitle(0, "Grid search maximisation of the loglikelihood");
	DrawLegend(0, 20, 140, 0);
	DrawX(0, mA[1][], mA[0][]);
	DrawAdjust(ADJ_SYMBOL, PL_CIRCLE, 4000/(cRange[2]));
	DrawAdjust(ADJ_SYMBOLUSE, 1);
	DrawLine(0, avP[0], mA[1][mB],  avP[0], mA[1][mC], 1 );
	DrawText(0, " Estimated $\\rho$ value", avP[0], (mA[1][mC]+(mA[1][mB]-mA[1][mC])*2/3),  -1 );
	DrawText(0, "$\\rho$ value", 0, mA[1][mB], -1 , -1 , TEXT_XLABEL);
	DrawText(0, "estimated loglikelihood", 0, mA[1][mB], -1 , -1 , TEXT_YLABEL);
	DrawX(0, adFunc[0], avP[0]);
	DrawAdjust(ADJ_SYMBOL, PL_FILLCIRCLE, 4000/(cRange[2]));
	DrawAdjust(ADJ_COLOR, 3);
	if (cPic)
	{
	ShowDrawWindow();
//	SaveDrawWindow("HillCLimb.eps");
	}
}	

HillClimb(const func, const avP, const adFunc, const cRange, const cPic) 
{
	// function, parameter to be changed, criterion
	// cRange. start, criterion
	decl i, mA, mB, mC, mD, mE, mF, mDir, mMstep, mStep, mAux, mZero;
	mC = zeros(3,1);
	mA = 1;
	mD = 1-2*(cRange[0]<0);		//outer limit (either -1 or 1)
	mMstep = min( (1-fabs(cRange[0]))/4, (-1+fabs(cRange[0]))/4)/1000; //make sure no zero or one crossing in intial step
	// find initial direction
	mStep = 1;

	mE = cRange[0]; // introduce moving starting value
	func( mE^s_mM, &mB, &mAux, 1); //initial step, round 1
	mC[0] = mB[0];
	func((mE+mStep*mMstep)^s_mM, &mB, &mAux, cPic); //initial step, round 1
	mDir = mC[0]-mB[0];	  // <0 if Loglik improved
	mDir = (-1)^(1+(mDir<0));
	// initial direction: down (mDir=-1), up (mDir=1)
	
	while(mA>cRange[1])			//convergence criterion
	{
	func((mE+mDir*mStep*mMstep)^s_mM, &mB, &mAux, cPic); //initial step, round 1
	mC[1] = mB[0];
	func((mE+mDir*3*mStep*mMstep)^s_mM, &mB, &mAux, cPic); //initial 2nd step, round 2
	mC[2] = mB[0];
		i = 3;
		while(i>1)  // likelihood improved?
		{
		//check if double step length is possible
			if( (fabs(mE+mDir*(2^i-1)*mMstep)>=1) || (mD*(mE+mDir*(2^i-1)*mMstep)<=0) )		   //too long a step? crossing 1
//			if( (fabs(mE+mDir*(2^i-1)*mMstep)>=1) )		   //too long a step? crossing 1
			{
			i -= 2;
			mC[0] = mC[1];
			mE += mDir*(2^i-1)*mMstep; //reset to last admissable value
			func((mE+mDir*mStep*mMstep)^s_mM, &mB, &mAux, cPic); //initial step, round 1
			mC[1] = mB[0];
			func((mE+mDir*3*mStep*mMstep)^s_mM, &mB, &mAux, cPic); //initial 2nd step, round 2
			mC[2] = mB[0];
			i = 3;
			}
		if (mC[1]>=mC[2]) 
			{
			mStep = 2^(i-2)-1;
			i = -1;
			}
		else if(mC[1]<mC[2])	//likelihood improved		 
			{
			mC[0] = mC[1];
			mC[1] = mC[2];
			func((mE+mDir*(2^i-1)*mMstep)^s_mM, &mB, &mAux, cPic);
			mC[2] = mB[0]; //new first step
			}
			++i;
		}
	//secure correct output
	avP[0] = mE+mDir*mStep*mMstep;
	adFunc[0] = mC[1];
	mE += mDir*mStep*mMstep; //take last value of parameter	to start from in opposite direction
	mA =  100*(fabs((mC[1]-mC[0])/mC[0])); // percentage improvement inloglik
	mStep = 1;	   //reset step length
	mMstep /= 2; //increase precision
	mC[0] = mC[1];
	}
}

MyCSeason(mT, mF) /* Calculates the centred seasonal dummies */
{
	//input: mT - length of vector, mF - frequency
	decl mA;
   mA = (1-1/mF[1])~(-1/mF[1])*ones(1,mF[1]-1);
   mA = mA'**(ones(mF[0]/mF[1],1));
   mA = (ones(ceil(mT/mF[0]+1),1)**mA);
   mA = lag0(mA[][0], range(0,(mF[0]/mF[1])*(mF[1]-1),mF[0]/mF[1])');
   return(mA[mF[0]:][0:mF[1]-2]);
}

Stopit()
{
decl i;
for (i=2; i>1; i++)
{
if (1)
continue;
}
}