#include <oxstd.h>					 
#include <oxfloat.h>
#include <oxdraw.h>					 
#import  <modelbase>
#import <lib/testres>

#include "MyArfima.h"

#ifdef NO_DLL
#include "fracsig.ox"
#endif

findsample(const mdata, const mvarsel, const mlagsel,
    const it1, const it2, const dmisval, const at1, const at2);

const decl DB_MISVAL = -9999.99;


/*-------------------------- Arfima : Database -----------------------------*/
Arfima::Arfima()
{
    this->Modelbase();
    SetMethod(M_MAXMPLIK);
	m_iModel = 0;

	FixMean(0);
    m_cZ = m_cW = 0;
	ARMA(0, 0);
	SetParCount(1);		// d parameter

    m_fDfixed = FALSE;
	m_mForc = m_mForcLevel = <>;
	m_cTdynforc = 8;

	m_asMethods = {"EML", "NLS", "MPL", "Init", "NLS-STAT"};
}

Arfima::GetPackageName()
{
	return "Arfima";
}
Arfima::GetPackageVersion()
{
	return "1.01";
}
Arfima::Arma(const cAR, const cMA)
{
    m_cAR = cAR;  m_cMA = cMA;
    SetModelStatus(MS_NONE);		       /* throw away current model/data */
	SetParCount(1 + cAR + cMA);
}
Arfima::ARMA(const cAR, const cMA)
{
	Arma(cAR, cMA);
}
Arfima::GetXNames()
{
	decl asylag = {}, asx = {};
	
	// exclude name of dep var here
	GetGroupLagNames(Y_VAR, 1, 10000, &asylag);
    GetGroupLagNames(X_VAR, 0, 10000, &asx);

	return asylag ~ asx;
}
Arfima::GetXBeta(const vP)
{
	if (m_cX)
		return m_mX * vP[1 + m_cAR + m_cMA : m_cAR + m_cMA + m_cX];
	else
		return 0;
}
Arfima::GetXBetaStart(const mY)
{
    if (m_cX)
    {
        decl vols;
		if (m_cW)
	        olsc(mY .* sqrt(m_mW), m_mX .* sqrt(m_mW), &vols);	     /* WLS */
		else
	        olsc(mY, m_mX, &vols);	  								 /* OLS */
		return vols';
	}
	return <>;
}		
Arfima::GetXSizeInit()
{
	return m_cX;
}
Arfima::GetXPartial()
{
	return TRUE;
}
Arfima::GetXBetaForc(const vP, const cTforc)
{
	decl mxf = GetForcData(X_VAR, cTforc);
	if (rows(mxf))
		return mxf * vP[1 + m_cAR + m_cMA : m_cAR + m_cMA + m_cX];
	else
		return <>;
}
Arfima::GetZNames()
{
	decl i, asz = {};

	// Z variables (if any)
	GetGroupLagNames(Z_VAR, 0, 10000, &asz);
	for (i = 0; i < sizeof(asz); ++i)
        asz[i] ~= " (Z)";
		
	return asz;
}
Arfima::GetZGamma(const vP)
{
	if (m_cZ)
		return m_mZ * vP[1 + m_cAR + m_cMA + m_cX : m_cAR + m_cMA + m_cX + m_cZ];
	else
		return 0;
}
Arfima::GetZGammaStart(const mY)
{
    if (m_cZ)
    {
        decl vols;
        olsc(mY, m_mZ, &vols);
		return vols';
	}
	return <>;
}		
Arfima::GetZSizeInit()
{
	return m_cZ;
}

Arfima::GetParNames()
{
	decl i, asp = {};

	// ARFIMA names
	asp ~= sprint("d parameter");
	for (i = 0; i < m_cAR; ++i)
        asp ~= sprint("AR-", "%-2d", i + 1);
    for (i = 0; i < m_cMA; ++i)
        asp ~= sprint("MA-", "%-2d", i + 1);
		
	return asp ~ GetXNames() ~ GetZNames();
}
Arfima::GetcDfLoss()
{
	return GetFreeParCount() + 1;				  /* includes variance term */
}
Arfima::GetForcData(const iGroup, const cTforc)
{
	decl mx, t1x, t1, t2, cx;
	decl t2selsave = m_iT2sel;					   /* save selection sample */
	
    m_iT2sel = rows(m_mData) - 1;		 /* full sample to get whole series */
	if (iGroup == Y_VAR)
    	mx = GetGroupLag(Y_VAR, 0, 0);
	else if (iGroup == X_VAR)
    	mx = GetGroupLag(Y_VAR, 1, 100) ~ GetGroup(iGroup);
	else
    	mx = GetGroup(iGroup);		
	m_iT2sel = t2selsave; 						  /* reset selection sample */
	
	t1x = t1 = m_iT2est + 1 - m_iT1sel;
	t2 = min(rows(mx) + m_iT1sel - m_iT1est - 1, t1 + cTforc - 1);
	cx = columns(mx);
	
	if (cx == 0 || t2 < t1)						 /* no post-estimation data */
		return <>;

    findsample(mx, range(0, cx-1), zeros(1, cx), t1, t2, DB_MISVAL, &t1, &t2);
	if (t1 == t1x && t2 >= t1)
	    return mx[t1 : t2][];
	else
		return <>;
}
Arfima::GetD()
{
    return m_vPar[0][0];
}
Arfima::FixD(const dD)
{
    m_fDfixed = TRUE;
	FixPar(0, dD);
}
Arfima::FixAR(const iOrder, const valAR)                    /* 1 to fix AR(1) term, etc. */
{
	decl i, ifix;
	for (i = 0; i < sizerc(iOrder); ++i)
	{
		ifix = iOrder[i];
		if (ifix > 0 && ifix <= m_cAR)
			FixPar(ifix, valAR[i]);
	}
}
Arfima::FixMA(const iOrder, const valMA)                    /* 1 to fix MA(1) term, etc. */
{
	decl i, ifix;
	for (i = 0; i < sizerc(iOrder); ++i)
	{
		ifix = iOrder[i];
		if (ifix > 0 && ifix <= m_cMA)
			FixPar(m_cAR + ifix, valMA[i]);
	}
}
Arfima::FreeD()
{
	FreePar(0);
    m_fDfixed = FALSE;
}
Arfima::FreeARMA()
{
	decl i;
	for (i = 1; i <= m_cAR + m_cMA; ++i)
		FreePar(i);
}
Arfima::FreeArma()
{
	FreeARMA();
}
Arfima::FixX()
{
	if (m_cX)										 /* fix beta's in theta */
	{	FixPar( range(1 + m_cAR + m_cMA, m_cAR + m_cMA + m_cX),
			m_vPar[1 + m_cAR + m_cMA : m_cAR + m_cMA + m_cX]);
	}
	return GetFreePar();
	
}
Arfima::FreeX()
{
	if (m_cX)										 /* free beta's in theta */
	{	
		FreePar(range(1 + m_cAR + m_cMA, m_cAR + m_cMA + m_cX));
		return m_vPar[1 + m_cAR + m_cMA : m_cAR + m_cMA + m_cX];
	}
	return <>;
}
Arfima::GetParStatus()
{
	return { m_fDfixed, m_cAR, m_cMA, m_vIsFreePar, m_vFixedPar,
		m_fKnownMean, m_dYmean, m_cX, m_cZ};
}
Arfima::GetFuncEval()
{
	return m_cFuncEval;
}
Arfima::SetFuncEval()
{
	m_cFuncEval = 0;
}
Arfima::FixMean(const dYmean)
{
    SetModelStatus(MS_NONE);
    m_dYmean = dYmean;
    m_fKnownMean = TRUE;
}
Arfima::SetKnownMean(const dYmean)
{
	FixMean(dYmean);
}
Arfima::GetMean()
{
    return m_dYmean;
}
Arfima::FreeMean()
{
	UseSampleMean();
}
Arfima::UseSampleMean()
{
    SetModelStatus(MS_NONE);
	m_dYmean = 0.0;
    m_fKnownMean = FALSE;
}
Arfima::SetSigma2(const dSigma)
{
    m_dLogVar = log(dSigma);
}
Arfima::GetSigma2()
{
    return exp(m_dLogVar);
}
Arfima::InitData()
{
    decl  cp, vp, mh, asy;

	if (!Modelbase::InitData())
		return FALSE;

	m_cFuncEval = 0;

	m_mZ = GetGroup(Z_VAR);
    m_cZ = columns(m_mZ);

    m_mW = GetGroup(W_VAR);
    m_cW = columns(m_mW);
	if (m_cW > 1)
		m_mW = m_mW[][0], m_cW = 1;
	else if (m_cW == 0)
		m_mW = 1, m_dLogWeight = 0;
		
    if (m_cW != 0)
    {
		decl mwold = m_mW;
		m_mW = m_mW .< 1e-10 .? 0 .: m_mW;	/* set small & neg. values to 0 */
		decl mw = double(meanc( m_mW[vecindex(m_mW .> 0)][] ));
		m_mW = m_mW / mw;			   /* scale by mean of positive weights */
		m_dLogWeight = double(sumc( log( m_mW[vecindex(m_mW .> 0)][] ) ));
    }
	
    if (!m_fKnownMean)
        m_dYmean = double(meanc(m_mY));                     /* mean of data */
    m_dYvar = double(varc(m_mY));                       /* variance of data */

    GetGroupLagNames(Y_VAR, 0, 0, &m_asY);
	m_asX = GetParNames();
	
    if (m_dYvar <= 1e-12)
    {   println("Data variance is less than 1e-12. This is too small.");
        return FALSE;
    }
    return TRUE;
}

static getPeriodogram(const mY, const iTrunc, const amFreq)
{
	decl ct, cp, cper, y, x;

	ct = rows(mY);
	cp = int(ct / 2);			 /* no. periodogram points excluding freq 0 */

	if (ct < 1024)			                        /* standard periodogram */
    {	y = periodogram(mY)[1:][]';					              /* in row */
		cper = ct;							     /* no of periodogram freqs */
	}
    else    /* use fft for large data sets (same answer if T is power of 2) */
    {   y = fft(mY');      	                 /* columns(y) is length of fft */
		cper = columns(y);						   /* no of fft frequencies */
      	y = y[][1 : cp];
      	y = sqr(cabs(y)) / (M_2PI * ct);
    }
	if (iTrunc > 0 && iTrunc < cp)
	{
		y = y[][:iTrunc-1];
		cp = iTrunc;
	}
	amFreq[0] = M_2PI * range(1, cp) / cper;/*frequencies of periodogram/fft*/
	return y;
}
Arfima::EstimateGPH(const mY, const iTrunc, const fPrint)
{
	decl y, x, d, sed, prd, dgph, vgph, mgph, vpos;

	y = getPeriodogram(mY, iTrunc, &x);/*y: periodogram; x: frequencies, 1xm*/

	vpos = vecindex(y .> 1e-8);    /* select non-zero values in periodogram */
    y = log(y[][vpos]);
    x = x[][vpos];
    x = 1 | -log(2*(1-cos(x)));

	olsr(y, x, &vgph, &mgph);	          /* Geweke Porter-Hudak regression */
	d = vgph[1];
	sed = M_PI * sqrt(mgph[1][1]/6);
    prd = 2 * tailn(fabs(d / sed));
												/* weights not supported !! */
	if (fPrint)
    {   print("\n----  Log Periodogram Regression ----\n");
        print("d parameter    " , "%14.6g", d,
            sprint(" (", sed, ")"), sprint(" [", "%.4f", prd, "]") );
		println("\nNo of observations: ", sizerc(mY),
			    "; no of periodogram points: ", sizerc(y));
    }
	return d ~ sed ~ prd;
}

Arfima::EstimateGSP(const mY, const iTrunc, const fPrint)
{
	decl y, x, meanlx, d, dd, yxpow2d, f, d1f, d2f, mean, i, sed, prd, mxiter, eps;

	y = getPeriodogram(mY, iTrunc, &x);/*y: periodogram; x: frequencies, 1xm*/
	meanlx = meanr(log(x));

	[mxiter] = GetMaxControl();
	[eps] = GetMaxControlEps();
	eps /= 100;
	
    for (i = 0, d = 0, dd = 1; i < mxiter && fabs(dd) > eps; i++)
	{
		yxpow2d = y .* (x .^ (2 * d));
		mean = meanr(yxpow2d);
		f = log(mean) - 2 * d * meanlx;
		d1f = 2 * meanr(yxpow2d .* log(x)) / mean - 2 * meanlx;
		d2f = 4 * meanr(yxpow2d .* sqr(log(x))) / mean;
		dd = double(d1f / d2f);
        d -= dd;
	}

	sed = 1 / (2 * sqrt(sizerc(y)));
    prd = 2 * tailn(fabs(d / sed));
	
	if (fPrint)
    {   print("\n----  Gaussian semiparametric estimate ----\n");
        print("d parameter    " , "%14.6g", d,
            sprint(" (", sed, ")"), sprint(" [", "%.4f", prd, "]") );
		println("\nNo of observations: ", sizerc(mY),
			    "; no of periodogram points: ", sizerc(y));
    }
	return d ~ sed ~ prd;
}

Arfima::SolveAR(const mY, const cP, const cQ, const mIsFree, const fPrint, ...)
{					   /* solve Yule-Walker equations (toeplitz if cQ == 0) */
	decl varg, macf, myw, vyw, vsel, vp, ck = 0, cpq, mxt;

	vp = zeros(1, cP);
	if (cP <= 0)
		return vp;
	vsel = vecindex(mIsFree .!= 0);

	varg = va_arglist();
	if (sizeof(varg))			
		ck = varg[0];			// number of extra terms for Yule-Walker eqns

	cpq = cP + cQ;
    macf = acf(mY, cpq + ck);					// get ACF
	macf = reversec(macf[1 : cP][]) | macf;     // add pre-sample values
	myw = lag0(macf, range(1,cP));				// regressors are lagged ACF
	mxt = ck ? rows(macf) - 1 : cpq + cP;		// last `observation'

	olsc(macf[cpq + 1 : mxt][], myw[cpq + 1 : mxt][vsel], &vyw);
	vp[][vsel] = vyw';							// store AR estimates

	if (fPrint)
    {   println("\n---- ", "ARMA(", cP, ",", cQ, ") ", 
			"solution to Yule-Walker equations ----");
        print("AR parameters" , "%14.6g", vp);
		print("T = ", rows(mY), "  free AR parameters: ", sizerc(vsel));
		if (ck)
			print("  extra ACF terms: ", ck);
		else
			print("\n");
    }
	return vp;
}
Arfima::SolveMA(const mY, const cQ, const mIsFree, const dItEps, const cItMax,
	const fPrint)
{
	decl i, j, cb, macf, vsel, vq, ma, vb, vbsel, vg, var, oldvar,
		v1, v2, cflip;

	vq = zeros(1, cQ);
	if (cQ <= 0)
		return vq;
	vsel = vecindex(mIsFree .!= 0);
		
            /* solve using Tunnicliffe-Wilson method (see e.g. Granger & Newbold, p.88) */
    var = double(varc(mY));

	cb = sizerc(vsel);
	vbsel = 0 | 1 + vsel;
    macf = acf(mY, cQ) * var;
    ma = new matrix[cQ+1][cQ+1];
    vb = new matrix[1][cQ+1];
    vb[0][0] = var;

    for (i = 0; i < cItMax; ++i)
    {	                       /* Note: this has very slow convergence rate */
        for (j = 0, vg = -macf'; j <= cQ; ++j)
        {
			v1 = vb[0][j:cQ];  v2 = vb[0][0:cQ-j];
            ma[j][] = v1;
            ma[j][j:] += v2;
            vg[0][j] += double(v1 * v2');
        }
        vb[0][vbsel] -= vg / ma[][vbsel]';
		          /* Note: <a1, a2, 0> not identical to <a1, a2>, but close */
		oldvar = var;
		var = vb[0][0]^2;
        if (i >= 2 && double(max(fabs(vg)) / var) < dItEps)
            break;
    }

    vq[0][] = vb[0][1:] / vb[0][0];
    if (cQ == 1 && fabs(vq[0][0]) > 1)
        vq[0][0] = 1 / vq[0][0];

	[vq, cflip] = FlipMA(vq);
	
	if (fPrint)
    {   println("\n----  ", "MA(", cQ, ") ", 
			"solution from Tunnicliffe-Wilson\'s method ----");
        print("MA parameters" , "%14.6g", vq');
		println("T = ", rows(mY), "  free MA params: ", sizerc(vsel),
				"  residual variance: ", var);
		if (cflip)
			println(cflip, " MA roots inverted");
    }
		
	return vq';
}

Arfima::FlipMA(const vMA)
{
	decl roots, vsel, cr;
	
	if (sizerc(vMA) == 0)
		return;
		
	polyroots(1 ~ vec(vMA)', &roots);	         // compute MA roots
	vsel = vecindex(cabs(roots) .> 1);
	cr = rows(vsel);	   // no of roots outside unit circle
	if (cr > 0)		      
	{					  // invert roots outside unit circle
		roots[][vsel] = cdiv(ones(1,cr), roots[][vsel]);
						        // reconstruct MA parameters
		return {polymake(roots)[0][1:]', cr};
	}
	return { vMA', 0 };
}

Arfima::InitPar()
{
    decl vp, i, j, y, x, cp, ct, d, mu, var;
	decl badjust = (m_iMethod != M_INITONLY);
	decl bprint = m_fPrint && m_iMethod == M_INITONLY;

	if (!Modelbase::InitPar())
		return FALSE;

	SetParCount(1 + m_cAR + m_cMA - m_cPar, 1);	 /* truncate to arfima pars */
	m_cX = GetXSizeInit();
	m_cZ = GetZSizeInit();
	SetParCount(m_cX + m_cZ, 1);			     /* then add regressor pars */

    vp = zeros(1, m_cPar);

    mu = m_mY - m_dYmean;             /* this will be sequentially filtered */

    if (m_cX)
    {
        vp[1+m_cAR+m_cMA : ] = GetXBetaStart(mu);
        mu -= GetXBeta(vp');
    }

    if (m_fDfixed)                                       /* yes: d is fixed */
        d = m_vFixedPar[0][0];
    else                                       /* find starting value for d */
    {	decl vd;
		vd = EstimateGPH(mu, m_cAR + m_cMA > 0 ? sqrt(m_cT) : m_cT, bprint);
		d = badjust && vd[2] > 0.05 ? 0 : vd[0];/* only if significant at 5%*/
    }
    if (badjust && d > 0.45)  d = 0.4;   /* restrict d to stationary values */
    if (badjust && d < -0.45) d = -0.4;  /* restrict d to stationary values */

	if (fabs(d) > 1e-4)
    {   mu = diffpow(mu, d);                   /* compute u_t = (1-L)^d y_t */
        mu[0][0] = 0;                            /* loses first observation */
    }
    vp[0] = d;
    if (m_cAR)                    /* find starting values for AR parameters */
    {            /* by solving Yule-Walker equations (toeplitz if m_cMA==0) */
		vp[1 : m_cAR] = SolveAR(mu, m_cAR, m_cMA, m_vIsFreePar[1 : m_cAR], bprint);
              /* ignore MA terms if starting values are outside unit circle */
		if (badjust && !IsStationary(vp[1 : m_cAR], 0))
			vp[1 : m_cAR] = SolveAR(mu, m_cAR, 0, m_vIsFreePar[1 : m_cAR], 0);
		
        mu = arma0(mu, vp[1:m_cAR], m_cAR, 0);	          /* take residuals */
	}

    if (m_cMA)
    {  /* find starting values for MA parameters by using T-Wilson's method */
		vp[m_cAR+1 : m_cAR+m_cMA] =
			SolveMA(mu, m_cMA, m_vIsFreePar[m_cAR+1 : m_cAR+m_cMA], 1e-3, 10, bprint);
        mu = arma0(mu, vp[1+m_cAR:m_cAR+m_cMA], 0, m_cMA);	   /* residuals */
    }
    if (m_cZ)	                              /* now regress residuals on Z */
    {
        vp[1+m_cAR+m_cMA+m_cX : ] = GetZGammaStart(mu);
        mu -= GetZGamma(vp');
    }
    var = double(varc(mu));
    m_dLogVar = log(var);                /* starting value for log(sigma^2) */
// println("Starting values for sigma2: ", exp(m_dLogVar));
	SetFreePar(MapParToFree(vp));
	ResetFixedPar();

    return TRUE;
}
Arfima::SetStartPar(const vParFree)
{
    if (!Modelbase::InitPar())
        return FALSE;
	SetParCount(1 + m_cAR + m_cMA - m_cPar, 1);	 /* truncate to arfima pars */
	m_cX = GetXSizeInit();
	m_cZ = GetZSizeInit();
	SetParCount(m_cX + m_cZ, 1);			     /* then add regressor pars */

	SetFreePar(vParFree);
    
    return TRUE;
}

Arfima::IsStationary(const mTheta, const dD)
{
    if (dD <= -1.0 || dD > 0.49999)				  /* infeasible range for d */
        return 0;

	if (sizerc(mTheta))         /* check if AR roots are inside unit circle */
    {   decl roots;
        polyroots(-1 ~ vec(mTheta)', &roots);
        if (max(cabs(roots)) >= 0.9999)               
            return 0;
	}
	return 1;
}

Arfima::Acf(const cT, const vPar, const cAR, const cMA, const bPrintErr)
{
    decl vsig, vp = vec(vPar)';	
	
	// compute autocorrrelations
	if (fabs(vp[0][0]) <= 1e-9)          /* d is 0: use standard evaluation */
    {                 /* Sigma is a Toeplitz matrix of the autocorrelations */
             /* note that when p=0, Sigma is banded toeplitz, bandwidth q+1 */
        if (cAR + cMA)
            vsig = armavar(vp[][1:], cAR, cMA, 1, cT);
        else
       		vsig = unit(1, cT);
    }
    else
	{	decl ierr = fracsigma(vp, cAR, cMA, 1.0, cT, &vsig);
		if (ierr != 0)
		{   if (bPrintErr)
    			print("fracsigma failed (code:", ierr, ")\n");
        	return 0;
		}
	}
	return vsig;
}
Arfima::Filter(const mY, const dYmean, const vPar, const cAR, const cMA,
	const mXBeta, const mZGamma, const adLogDet, ...)
{
	decl ct = rows(mY), cy = columns(mY), bprint = TRUE, logdet;
	decl va = va_arglist();
	if (sizeof(va) > 0)
		bprint = va[0];

	decl vsig = Acf(ct, vPar, cAR, cMA, bprint);			  /* ARFIMA ACF */

	if (ct != columns(vsig))								  /* ACF failed */
		return 0;
	
    decl vy = mY - dYmean;						/* compute mean-corrected y */
	if (columns(mXBeta))	 /* remove additive regressors before filtering */
        vy -= mXBeta;

	vy = pacf(vsig', &logdet, vy);			  /* apply filter to demeaned y */
	if (vy == 0)
		return 0;							 				 /* pacf failed */
	vy = vy[][ : cy - 1];

    if (columns(mZGamma))		   /* finally, remove innovative regressors */
    	vy -= mZGamma;

	if (isarray(adLogDet))
		adLogDet[0] = logdet;
	
	return vy;
}
Arfima::FilterNaive(const mY, const dYmean,	const vPar, const cAR, const cMA,
	const mXBeta, const mZGamma, const adLogDet, ...)
{
    decl d, vp = vec(vPar)';
	
    decl vy = mY - dYmean;						/* compute mean-corrected y */
	if (columns(mXBeta))	 /* remove additive regressors before filtering */
        vy -= mXBeta;

	d = vp[0][0];
    if (d != 0)
    {
		if (d < -100) d = -100;
		vy = diffpow(vy, d > 1000 ? 1000 : d); /* compute u_t = (1-L)^d y_t */
        vy[0][0] = 0;
    }
    if (cAR + cMA)                       /* take residuals from arma filter */
        vy = arma0(vy, vp[0][1 : ], cAR, cMA);

    if (columns(mZGamma))		   /* finally, remove innovative regressors */
    	vy -= mZGamma;

	if (isarray(adLogDet))
		adLogDet[0] = 0;				     /* always zero in naive filter */
		
	return vy;
}

Arfima::fLogLik(const vP, const adFunc, const avScore, const amHessian)
{
    decl ve, mv, logdet, drss, logdetx = 0, d;

	SetFreePar(vP);
	++m_cFuncEval;
	d = m_vPar[0][0];

	if (!Arfima::IsStationary(!m_cAR || fabs(d) <= 1e-9 ? <> : m_vPar[1 : m_cAR], d))
		return 0;			  			       /* d not 0: Filter checks AR */
	
	if (m_bConcentrateX)	// concentrating out X's: filter them
		ve = Filter(m_mY - m_dYmean ~ m_mX, 0, m_vPar, m_cAR, m_cMA,
				<>,	GetZGamma(m_vPar), &logdet, m_fPrint);
	else					// else subtract Xbeta
		ve = Filter(m_mY, m_dYmean, m_vPar, m_cAR, m_cMA,
				GetXBeta(m_vPar), GetZGamma(m_vPar), &logdet, m_fPrint);
	
	if (ve == 0)
		return 0;

	if (m_bConcentrateX)
		mv = ve[][1:];										/* filtered X's */
	ve = ve[][0];											  /* filtered Y */
		
	if (m_bConcentrateX)			   /* concentrate X's out of likelihood */
	{
		decl mvinv = invertsym(mv'mv, &logdetx);  	  /* store log variance */
		decl mbeta = (ve'mv) * mvinv;		  /* concentrated beta (in row) */
		m_vPar[1 + m_cAR + m_cMA : ] = mbeta;	 	          /* store beta */			
		drss = double(ve've - mbeta * (mv've));
	}
	else if (m_cW)
	{	logdet -= m_dLogWeight;
		drss = double(ve' (ve .* m_mW));				    /* weighted RSS */
	}
	else
		drss = double(ve've);
	
	if (m_iMethod == M_MAXMPLIK)
	{		   /* modified profile log-likelihood, An and Bloomfield (1993) */
		decl cti = 1 / m_cT;
		m_dLogVar = log(drss / (m_cT - m_cX));
		adFunc[0] =	((cti - 0.5) * logdet - 0.5 * logdetx +
			(1 + 0.5 * (m_cX - m_cT)) * log(drss * cti)) * cti;
	}
	else										 /* standard log-likelihood */
	{	m_dLogVar = log(drss / m_cT);
		adFunc[0] = -0.5 * (logdet / m_cT + m_dLogVar);
	}

return 1;
}

Arfima::fNLS(const vP, const adFunc, const avScore, const amHessian)
{
    SetFreePar(vP);
	++m_cFuncEval;

    if (m_iMethod == M_NLS_STATIONARY)
    {
		if (m_cAR && !Arfima::IsStationary(m_vPar[1 : m_cAR], 0))
			return 0;
    }
	decl ve = FilterNaive(m_mY, m_dYmean, m_vPar, m_cAR, m_cMA,
		GetXBeta(m_vPar), GetZGamma(m_vPar), 0);

	decl drss;
	if (m_cW)
		drss = double(ve' (ve .* m_mW));
	else
		drss = double(sumsqrc(ve));
	// compute store log variance
	m_dLogVar = log(drss / (m_cT - m_cX));
		
    adFunc[0] = -0.5 * log(drss / m_cT);

return 1;
}

Arfima::DoEstimation(vPar)
{
    decl  cp = sizerc(vPar);

	m_mForc = m_mForcLevel = <>;
	m_bConcentrateX = FALSE;

	if (m_iMethod == M_INITONLY)		    /* only compute starting values */
    {
		m_iResult = MAX_CONV;
		m_dLogLik = M_NAN;
        fNLS(vPar, &m_dLogLik, 0, 0);
		m_mResidual = GetResiduals();
		return vPar;								   /* no further output */
	}
    else if (cp == 0)					       /* no parameters to estimate */
    {
		m_iResult = MAX_CONV;
        if (m_iMethod == M_NLS || m_iMethod == M_NLS_STATIONARY)
            fNLS(vPar, &m_dLogLik, 0, 0);
		else
            fLogLik(vPar, &m_dLogLik, 0, 0);
    }
	else
    {
		if (m_iMethod == M_MAXMPLIK && (m_cZ || m_cW))
		{	println("Z or W variables in model: M_MAXLIK replaces M_MAXMPLIK");
			m_iMethod = M_MAXLIK;
		}

		m_bConcentrateX = m_cX > 0 && m_cZ == 0 && m_cW == 0 && GetXPartial();

		if (m_iMethod == M_MAXLIK || m_iMethod == M_MAXMPLIK)
		{
			if (m_bConcentrateX)
				FixX();		   /* remove beta's from params to iterate over */
			decl vp = GetFreePar();
			m_iResult = MaxBFGS(fLogLik, &vp, &m_dLogLik, 0, TRUE);
			if (m_bConcentrateX)
				vp |= FreeX();		 /* concentrated X: add beta back to vp */
			vPar = vp;
		}
        else if (m_iMethod == M_NLS || m_iMethod == M_NLS_STATIONARY)
            m_iResult = MaxBFGS(fNLS, &vPar, &m_dLogLik, 0, TRUE);
    }
    m_dLogLik -= (1 + log(M_2PI)) / 2;	 /* adjust for scaling and constant */
    m_dLogLik *= m_cT;

	if (m_iResult == MAX_CONV || m_iResult == MAX_WEAK_CONV)
		m_iModel++;
	
    if (m_cMA)					/* reciprocal of MA roots if non-invertable */
	{	decl cflipma;
		SetFreePar(vPar);
		decl vp = GetPar();
		[ vp[1+m_cAR : m_cAR+m_cMA], cflipma ] = FlipMA(vp[1+m_cAR : m_cAR+m_cMA]);
		SetPar(vp);
		ResetFixedPar();
		vPar = GetFreePar();
	}

	m_mResidual = GetResiduals();

	return { vPar, "BFGS", TRUE };
}
Arfima::Covar()
{
    decl covar, vpsave = m_vPar, logvarsave = m_dLogVar, result, vp;

	m_mCovar = <>;
	
	if (m_iMethod == M_MAXMPLIK || m_iMethod == M_MAXLIK)
	{
		decl covarma = <>, covbeta = <>;
		// 2nd der. of modified profile log-likelihood
		if (m_bConcentrateX)
			FixX();		 /* remove beta's from params to differentiate over */
		vp = GetFreePar();
		if (sizerc(vp))
		{	result = Num2Derivative(fLogLik, vp, &covarma);
			covarma = invertgen(-covarma, 30) / m_cT;
		}
		if (m_bConcentrateX)
			FreeX();									 /* free them again */
		
		if (m_bConcentrateX)/*GLS-style var(Beta), assumes block diagonality*/
		{	decl mv, mvinv;
	                   /* apply the filter to X, returning V, X'inv(R)X=V'V */
			mv = Filter(m_mX, 0, m_vPar, m_cAR, m_cMA, <>, <>, 0);
			covbeta = invertgen(mv'mv / exp(logvarsave), 30);   /* variance */
		}
		decl ca = rows(covarma), cb = rows(covbeta);
		m_mCovar = zeros(ca + cb, ca + cb);
		if (ca)	m_mCovar[][] = covarma;
		if (cb)	m_mCovar[ca:][ca:] = covbeta;
	}
	else
	{
		vp = GetFreePar();
		if (sizerc(vp))
		{
			result = Num2Derivative(fNLS, vp, &covar);
	    	m_mCovar = invertgen(-covar, 30) / m_cT; /* fLogLik scales by T */
		}
	}

	m_vPar = vpsave;	       /* restore: Num2Der will have changed m_vPar */
    m_dLogVar = logvarsave;                             /* and log variance */

	if (rows(m_mCovar))
	{	decl mstderr = zeros(1, sizerc(m_vPar));
	    mstderr[][vecindex(m_vIsFreePar)] = sqrt(diagonal(m_mCovar));
		return mstderr';
	}

	return <>;
}

Arfima::CumForc(const dYlevel, mYforc, mYact, mMaCoefs, const dSigma2,
	mYforcOpt, mForcVar)
{
    mMaCoefs = (cumulate(mMaCoefs'))';
                                                 /* translate all to levels */
	mYforc = dYlevel + cumulate(mYforc);
	decl mrmse = sqrt(cumulate(dSigma2 .* sqr(mMaCoefs')));
	
	if (rows(mYact))
		mYact = dYlevel + cumulate(mYact);

	decl cfopt = rows(mYforcOpt);
	decl mrmseopt = 0;
	if (cfopt)
	{
	    mYforcOpt = dYlevel + cumulate(mYforcOpt);
	    // compute variance of cumsums of forecast errors using their
	    //       covariance matrix as computed for differenced series:
		decl mcum = lower(ones(cfopt, cfopt));
		mForcVar = mcum * mForcVar * mcum';
		mrmseopt =	sqrt(dSigma2 * diagonal(mForcVar)');
	}
	return {mrmse, mrmseopt, mYforc, mYact, mMaCoefs, mYforcOpt, mForcVar};
}

Arfima::Forecast(const cTforc, ...)
{
    decl k, vsig, mx, mxbeta, mxforc, mxbforc, myact, mbeta, mgamma, mforc, mrmse, mrmsen;
    decl i, t1, t2, ct, ctf, bnaiveonly, va, vylevel = <>, dexp = 0;
	decl fprint = m_fPrint;
	decl sigma2 = exp(m_dLogVar);

	m_mForc = m_mForcLevel = m_mForcExp = <>;

	if (cTforc <= 0)
		return;
    if (m_cZ)
    {   println("\n*** Forecasting with Z variables not implemented!");
        return 0;
    }

    if (m_iModelStatus < MS_DATA)
        InitData();
	else
	{	// if necessary: grow database, recreating deterministic variables
		Grow(cTforc + m_iT2est + 1 - rows(m_mData));
	}	
	bnaiveonly = FALSE;
	va = va_arglist();
	if (sizeof(va) > 0)
		vylevel = matrix(va[0]);
	if (sizeof(va) > 1)
		bnaiveonly = va[1];
	if (sizeof(va) > 2 && sizerc(va[2]))
		sigma2 = va[2];
	if (sizeof(va) > 3)
		fprint = int(va[3]);
	if (sizeof(va) > 4)
		dexp = va[4];
		
    if (m_cW && fprint)
        println("NB. Forecasting sets weights to unity");
		
	/*---------------------- get XBeta for forecast period -----------------*/
	mxforc = mxbeta = m_dYmean;
    if (m_cX)
    {   mxbforc = GetXBetaForc(m_vPar, cTforc);	 /* Xb over forecast period */
		ctf = min(rows(mxbforc), cTforc);

		if (ctf < cTforc)
        {   println("*** Error: Need regressors over full forecast period!");
            println("Available sample=", ctf, " requested no of forecasts=", cTforc);
            return 0;
        }                         /* get Xb over fitted and forecast period */
        mxbeta += GetXBeta(m_vPar);									  /* Xb */
        mxforc += mxbforc;
    }
	/*------------------------ get y for forecast period -------------------*/
	myact = GetForcData(Y_VAR, cTforc);
    ctf = min(rows(myact), cTforc); /* no of known y's over forecast period */
    myact = ctf > 0 ? myact[ : ctf - 1][] : <>;
	myact |= constant(M_NAN, cTforc - ctf, 1);  /* fill with missing values */

	/*------------------------ AR representation ---------------------------*/
                                                      /* ar coefs of F part */
    decl arcoefs = diffpow(1 | zeros(m_cT + cTforc - 1,1), m_vPar[0])';
    if (m_cAR)                                       /* ar coefs of AR part */
        arcoefs = polymul(arcoefs, 1 ~ -m_vPar[1 : m_cAR]')[][:m_cT+cTforc-1];
    if (m_cMA)                                       /* ar coefs of MA part */
        arcoefs = polydiv(arcoefs, 1 ~ m_vPar[1+m_cAR : m_cAR+m_cMA]',
            m_cT+cTforc);
    decl macoefs = polydiv(<1>, arcoefs[0][0:cTforc - 1], cTforc);
    arcoefs = -reverser(arcoefs[][1:]);

	/*--------------------- construct naive forecasts ----------------------*/
    decl vyf = (m_mY - mxbeta) | zeros(cTforc, 1);

    for (i = 0; i < cTforc; ++i)
         vyf[m_cT + i][0] = arcoefs[0][cTforc - 1 - i:] * vyf[:m_cT + i - 1][0];

    decl mnaive = vyf[m_cT:][] + mxforc;                  /* adjust for X's */

    mrmsen = sqrt(sigma2 * cumulate(sqr(macoefs')));

	/*------------------ construct best linear forecasts -------------------*/
	mforc = mrmse = <>;
	if (!bnaiveonly)
    {
		vsig = Acf(m_cT + cTforc, m_vPar, m_cAR, m_cMA, fprint);
		if (columns(vsig) == 0)
        {	print("*** Warning: Forecast(): ACF failed, only reporting naive forecasts!\n");
			bnaiveonly = TRUE;
		}
	}
	if (!bnaiveonly)
    {
        mgamma = new matrix[1][0];             /* do all forecasts together */
        for (k = cTforc - 1; k >= 0; --k)       /* gamma_k^{n}, Beran p.165 */
            mgamma ~= vsig[][cTforc-k : m_cT+cTforc-1-k]';
        mgamma = reversec(mgamma);
	    
        vsig = vsig[][:m_cT-1];                                 /* Sigma _n */
               /* we need beta = inv(S)*gamma, solve gamma=S*beta, for beta */
        mbeta = solvetoeplitz(vsig, m_cT, mgamma);
        if (columns(mbeta) == 0)
        {   print("*** Warning: Forecast(): solvetoeplitz failed!\n");
            return 0;
        }
        mforc = (mbeta'(m_mY - mxbeta)) + mxforc;       /* normal forecasts */
                 /* we also need gamma'*inv(S)*gamma, so compute gamma'beta */
		mrmse =	sqrt(exp(m_dLogVar)) * sqrt((vsig[0][0] - diagonal(mgamma'mbeta)'));
	}

	m_mForc = bnaiveonly ? mnaive : mforc;
	m_mForc ~= bnaiveonly ? mrmsen : mrmse;
	m_mForc ~= myact;

	if (fprint)
    {
        print("Forecasts from ", ObsYear(m_iT2est + 1), "(", ObsPeriod(m_iT2est + 1), ")");
		if (bnaiveonly)
			print("%c", {"Horizon", "naive forc", "(SE)", "actual"},
        	    range(1,cTforc)' ~ mnaive ~ mrmsen ~ myact);
		else
			print("%c", {"Horizon", "forecasts", "(SE)", "naive forc", "(SE)", "actual"},
        	    range(1,cTforc)' ~ mforc ~ mrmse ~ mnaive ~ mrmsen ~ myact);
    }
	// built up return matrix
	mx = mnaive ~ mrmsen;
	decl me = <>;
    if (cTforc == ctf)			// output actual forecast errors if computable
		me = (mnaive - myact);

	if (!bnaiveonly)
	{
    	mx = mforc ~ mrmse ~ mx;
	    if (cTforc == ctf)			// output actual forecast errors if computable
		 	me = (mforc - myact) ~ me;
	}
	mx ~= me;

	decl cdiff = sizerc(matrix(vylevel));
	if (cdiff)
    {
		decl vyinit = zeros(1, cdiff);
		vylevel = vec(vylevel);
		for (i = 0; i < cdiff; ++i)
		{
			vyinit[0][cdiff - i - 1] = vylevel[cdiff - 1][0];
			vylevel = diff0(vylevel, 1);
		}

		decl mforcvar = <>;
		if (!bnaiveonly)
			mforcvar = toeplitz(vsig[][:cTforc-1]) - mgamma'mbeta;
			
		for (i = 0; i < cdiff; ++i)
			[mrmsen, mrmse, mnaive, myact, macoefs, mforc, mforcvar] =
				CumForc(vyinit[0][i], mnaive, myact, macoefs,
					sigma2,	mforc, mforcvar);

		decl mxl = mnaive ~ mrmsen;
	    if (cTforc == ctf)
			me = (mnaive - myact);
		if (!bnaiveonly)
		{
    		mxl = mforc ~ mrmse ~ mxl;
		    if (cTforc == ctf)
			 	me = (mforc - myact) ~ me;
		}
		mx ~= mxl ~ me;

		if (fprint)
        {
            print("\nLevel forecasts, i.e. for the ",
				cdiff > 1 ? sprint(cdiff,"x ") : "", "cumulated variable:");
			if (bnaiveonly)
    	        print("%c", {"Horizon", "naive forc", "(SE)", "actual"},
        	        range(1,cTforc)' ~ mnaive ~ mrmsen ~ myact);
			else
    	        print("%c", {"Horizon", "forecasts", "(SE)" , "naive forc", "RMSE", "actual"},
        	        range(1,cTforc)' ~ mforc ~ mrmse ~ mnaive ~ mrmsen ~ myact);
        }
		m_mForcLevel = bnaiveonly ? mnaive : mforc;
		m_mForcLevel ~= bnaiveonly ? mrmsen : mrmse;
		m_mForcLevel ~= myact;
    }
	if (dexp)
	{
		decl mf = cdiff ? m_mForcLevel : m_mForc;
		decl mlo = exp(mf[][0] - dexp * mf[][1]);
		decl mhi = exp(mf[][0] + dexp * mf[][1]);
		decl ef = exp(mf[][0] + 0.5 * mf[][1] .^ 2);
		m_mForcExp = ef ~ mlo ~ mhi ~ exp(mf[][2]);

		if (fprint)
        {
            print("\nExponentiated ", cdiff ? "Level " : "", "forecasts (with ",
				dexp, "SE bands):");
   	        print("%c", {"Horizon", "forecasts", "lower", "upper", "actual"},
       	        range(1,cTforc)' ~ m_mForcExp);
		}
	}
	
	if (fprint)
		print("\n");
		
return mx;
}
Arfima::GetResiduals()
{
	decl v;
	if (m_iMethod == M_MAXLIK || m_iMethod == M_MAXMPLIK)
	{
		v = Filter(m_mY, m_dYmean, m_vPar, m_cAR, m_cMA,
			GetXBeta(m_vPar), GetZGamma(m_vPar), 0);
	}
	else
	{
		v =	FilterNaive(m_mY, m_dYmean, m_vPar, m_cAR, m_cMA,
			GetXBeta(m_vPar), GetZGamma(m_vPar), 0);
	}
	
	if (m_cW)
		v = v .* sqrt(m_mW);						  /* weighted residuals */
		
	return v;
}
Arfima::GetNaiveResiduals()
{
	decl v;
	
	v = FilterNaive(m_mY, m_dYmean, m_vPar, m_cAR, m_cMA,
		GetXBeta(m_vPar), GetZGamma(m_vPar), 0);

	if (m_cW)
		v = v .* sqrt(m_mW);						  /* weighted residuals */
		
	return v;
}
Arfima::GetModelLabel()
{
	return sprint("Arfima(", "%2d", m_iModel, ")");
}
Arfima::GetMethodLabel()
{
	return m_asMethods[m_iMethod];
}
Arfima::Output()
{
    decl i, j, k, csel, asy, asx, asz, cp, stitle;

    GetGroupLagNames(Y_VAR, 0, 0, &asy);

    if (m_cX)
        GetGroupLagNames(X_VAR, 0, 10000, &asx);
    if (m_cZ)
        GetGroupLagNames(Z_VAR, 0, 10000, &asz);

    stitle = sprint(
		m_cW ? "Weighted " : "",
        (m_iMethod == M_MAXLIK) ? "Maximum likelihood" :
        (m_iMethod == M_MAXMPLIK) ? "Modified profile likelihood" :
        (m_iMethod == M_NLS) ? "Non-linear least squares" :
        (m_iMethod == M_NLS_STATIONARY) ? "NLS (with stationarity)" :
        (m_iMethod == M_INITONLY) ? "Starting values" : "??",
        " estimation of ARFIMA(",
        m_cAR, m_fDfixed ? ",0," : ",d,", m_cMA, ") model");

	if (!OutputHeader(stitle))		// returns FALSE if no estimation
		return;

    if (!m_fKnownMean)
        println("(in deviation from sample mean)");
    else if (m_dYmean != 0)
        println("(in deviation from imposed mean ", m_dYmean, ")");
	print("\n");

	OutputPar();
	OutputLogLik();
    println("sigma", "%24.6g", sqrt(exp(m_dLogVar)), "  sigma^2", "%22.6g", exp(m_dLogVar));
    if (m_iMethod == M_INITONLY)
		println("(log-likelihood and standard errors based on NLS function)");
    print("\n");

	if (m_cW)
	{
		decl asw, cw0;
		GetGroupLagNames(W_VAR, 0, 0, &asw);
		println("\nWeighted estimation, W=", asw[0], " (scaled by Arfima package)");
    	println("mean(W): 1",
		        "  var(W): ", double(varc(m_mW[vecindex(m_mW .> 0)][])));
		decl dmy = double(sumc(m_mY .* sqrt(m_mW))) / double(sumc(sqrt(m_mW)));
		println("weighted mean(Y) = ", dmy,
			    "  weighted var(Y) = ", double(sumsqrc((m_mY .* sqrt(m_mW)) - dmy) / m_cT));
		cw0 = int(sumc(m_mW .== 0));
		if (cw0)
			println(cw0, " observation(s) with zero weight innovations");
	}
}
Arfima::TestGraphicAnalysis()
{
	decl vres = m_mResidual, asy;
    GetGroupLagNames(Y_VAR, 0, 0, &asy);

	DrawTMatrix(0, (m_mY-vres)' | m_mY', {"Fitted"} ~ asy, ObsYear(m_iT1est), ObsPeriod(m_iT1est), GetFrequency());
	DrawTMatrix(1, vres', "Residuals", ObsYear(m_iT1est), ObsPeriod(m_iT1est), GetFrequency());		
	DrawAdjust(ADJ_INDEX, 1, 0);
	DrawMatrix(2, acf(vres, min(100, rows(vres)))[1:]', "ACF-Residuals", 1, 1);		
	DrawAdjust(ADJ_INDEX, 1, 0);
	DrawAdjust(ADJ_MINMAX, -1, 1);
	ShowDrawWindow();
}
Arfima::TestForecastGraphics(const bExp)
{
	decl iplot = 0, mf, s;
	decl year1 = ObsYear(m_iT2est + 1), per1 = ObsPeriod(m_iT2est + 1), freq = GetFrequency();
	if (rows(m_mForc) > 0)
	{
		s = "Forecasts";
		DrawTMatrix(iplot, m_mForc[][0]', s, year1, per1, freq, 0, 2);
		DrawZ(m_mForc[][1]');
		DrawTMatrix(iplot++, m_mForc[][2]', "Actual", year1, per1, freq, 0, 3);
		mf = m_mForc;
	}
	if (rows(m_mForcLevel) > 0)
	{
		s = "Level Forecasts";
		DrawTMatrix(iplot, m_mForcLevel[][0]', s, year1, per1, freq, 0, 2);
		DrawZ(m_mForcLevel[][1]');
		DrawTMatrix(iplot++, m_mForcLevel[][2]', "Actual", year1, per1, freq, 0, 3);
		mf = m_mForcLevel;
	}
	if (bExp)
	{
		if (rows(mf) > 0)
		{
			decl mlo = exp(mf[][0] - 2*mf[][1]);
			decl mhi = exp(mf[][0] + 2*mf[][1]);
			mf[][0] = exp(mf[][0] + 0.5*mf[][1].^2);
			mf[][2] = exp(mf[][2]);
			DrawTMatrix(iplot, mf[][<0,2>]', {sprint("exp(", s, ")"), "Actual"}, year1, per1, freq);
			DrawTMatrix(iplot, mlo', "", year1, per1, freq, 0, 4);
			DrawTMatrix(iplot, mhi', "", year1, per1, freq, 0, 4);
			iplot++;
		}
	}
	if (iplot)
		ShowDrawWindow();
}
Arfima::TestSummary()
{
	Buffering(1);
	
	println("Descriptive statistics for residuals:");
	::NormalityTest(m_mResidual, TRUE);
	::ArchTest(m_mResidual, 1, 1, GetFreeParCount() - 1, TRUE);

	decl cparma = m_cAR + m_cMA ? int(sumc(m_vIsFreePar[1 : m_cAR + m_cMA])) : 0;
	if (!m_fDfixed) ++cparma;
	::PortmanteauTest(m_mResidual, max(int(sqrt(rows(m_mResidual))),
		3 * GetFrequency()), cparma, TRUE);

	Buffering(0);	
}
/*------------------------ END Arfima : Database ---------------------------*/
///////////////////////////////////////////////////////////////////////
// OxPack related
Arfima::SendVarStatus()
{
    return
        {{ "&Y endogenous", 'Y', STATUS_GROUP + STATUS_ENDOGENOUS, Y_VAR},
         { "&X variable", 'X', STATUS_GROUP, X_VAR},
         { "&Z variable", 'Z', STATUS_GROUP, Z_VAR},
         { "&Weight",     'W', STATUS_GROUP + STATUS_ONEONLY, W_VAR}};
}
Arfima::ReceiveModel()
{
    // get selection of database variables
    Select(Z_VAR, "OxPackGetData"("SelGroup", Z_VAR));
    Select(W_VAR, "OxPackGetData"("SelGroup", W_VAR));
	Modelbase::ReceiveModel();
}
Arfima::GetSettingValues()
{
	decl fixar, fixma, fixd, imean, dmean;
	fixar = m_cAR ? 1 + vecindex(!m_vIsFreePar[1 : m_cAR]) : <>;
	fixma = m_cMA ? 1 + vecindex(!m_vIsFreePar[m_cAR + 1 : m_cAR + m_cMA]) : <>;
	fixd = m_fDfixed && sizerc(m_vFixedPar) ? m_vFixedPar[0] : 0;
    imean = 0;
	if (!m_fKnownMean)       imean = 1;
    else if (m_dYmean != 0)	 imean = 2;
	dmean = imean == 3 ? GetMean() : 0;
	return {fixar, fixma, fixd, imean, dmean};
}
Arfima::SendDialog(const sDialog)
{
	if (sDialog == "OP_SETTINGS")
	{
		decl fixar, fixma, fixd, imean, dmean;
		[fixar, fixma, fixd, imean, dmean] = GetSettingValues();
		
		return
		{
			{ "ARMA" },
			{ "AR order", CTL_INT, m_cAR, "p" },
			{ "MA order", CTL_INT, m_cMA, "q" },
			{ "Fix AR lags", CTL_STRMAT, fixar, "fixp" },
			{ "Fix MA lags", CTL_STRMAT, fixma, "fixq" },
			{ "Fractional parameter d"},
			{ "Estimate d",	CTL_RADIO, m_fDfixed, "estd"},
			{ "Fix d",	    CTL_RADIO},
			{ "at:", CTL_DOUBLE, fixd, "fixd" },
			{ "Treatment of mean"},
			{ "None (or using Constant as regressor)", 	CTL_RADIO, imean, "mean"},
			{ "Deviation from sample mean", 			CTL_RADIO},
			{ "Fix mean",	    						CTL_RADIO},
			{ "at:", CTL_DOUBLE, dmean, "fixm" }
		};
	}
	// allow base class to process unhandled cases
	return Modelbase::SendDialog(sDialog);
}
Arfima::SendMethods()
{
	return
		{{ "Maximum Likelihood", 				M_MAXLIK, FALSE, 0},
		 { "Non-linear Least Squares",			M_NLS, FALSE, 0},
		 { "Modified Profile Likelihood", 		M_MAXMPLIK, FALSE, 0},
		 { "Starting values only",				M_INITONLY, FALSE, 0},
		 { "NLS with stationarity imposed", 	M_NLS_STATIONARY, FALSE, 0}
		};
}
Arfima::SendMenu(const sMenu)
{
	if (sMenu == "ModelClass")
	{
		return 
			{{ "&1: Arfima Models", 		"modelclass0", 1}};
	}
	else if (sMenu == "Test")
	{	return 
			{{ "&Graphic Analysis", "OP_TEST_GRAPHICS"},
			 { "&Forecast...", "OP_TEST_FORECAST"},
			 0,
			 { "&Test Summary", "OP_TEST_SUMMARY"},
			 0,
			 { "&Exclusion Restrictions...", "OP_TEST_SUBSET"},
			 { "&Linear Restrictions...",    "OP_TEST_LINRES"},
			 0,
			 { "Store in D&atabase...", "OP_TEST_STORE"}
			 };
	}
}
Arfima::DoStore(const sOpt, const fQuery)
{
	if (sOpt == "residuals")
		"OxPackStore"(m_mResidual, m_iT1est, m_iT2est, m_cW ? "WResidual" : "Residual", fQuery);
	else if (sOpt == "forecasts")
		"OxPackStore"(m_mForc[][0], m_iT2est + 1, m_iT2est + rows(m_mForc), "Forecast", fQuery);
	else if (sOpt == "forcse")
		"OxPackStore"(m_mForc[][1], m_iT2est + 1, m_iT2est + rows(m_mForc), "ForcSE", fQuery);
	else if (sOpt == "forclevel")
		"OxPackStore"(m_mForcLevel[][0], m_iT2est + 1, m_iT2est + rows(m_mForc), "ForcLevel", fQuery);
	else if (sOpt == "forclevelse")
		"OxPackStore"(m_mForcLevel[][1], m_iT2est + 1, m_iT2est + rows(m_mForc), "ForcLevelSE", fQuery);
	else
		return 0;
	return 1;
}

Arfima::ReceiveDialog(const sDialog, const asOptions, const aValues)
{
	if (sDialog == "OP_SETTINGS")
	{
		ARMA(aValues[0], aValues[1]);
		if (sizerc(aValues[2]) > 0)
			FixAR(aValues[2]);
		if (sizerc(aValues[3]) > 0)
			FixMA(aValues[3]);
		if (aValues[4] == 0)
			FreeD();
		else
			FixD(aValues[5]);

		if (aValues[6] == 2)
			FixMean(aValues[7]);
		else if (aValues[6] == 1)
			UseSampleMean();
		else
			FixMean(0);
		return 1;
	}
	else if (sDialog == "OP_TEST_FORECAST")
	{
		decl asoptions, avalues;
		if ("OxPackDialog"("Forecast",
				{	{ "Forecasting" },
					{ "Number of forecasts", CTL_INT, GetFrequency(), "cforc" },
					{ "Naive forecasts only", CTL_CHECK, 0, "naive" },
					{ "To undo s^th difference, specify s initial values for levels:" },
					{ "Base level for re-integration", CTL_STRMAT, <>, "base" },
					{ "Undo logarithm", CTL_CHECK, 0, "exp" },
				 	{ "Graphics" },
					{ "Graph forecasts", CTL_CHECK, 1, "graph" }
				},
				&asoptions, &avalues)
			)
		{
			Forecast(avalues[0], avalues[2], avalues[1]);
			if (avalues[4] > 0)
				TestForecastGraphics(avalues[3]);
		}
	}
	else if (sDialog == "OP_TEST_GRAPHICS")
		TestGraphicAnalysis();
	else if (sDialog == "OP_TEST_SUMMARY")
	{
		TestSummary();
	}
	else if (sDialog == "OP_TEST_STORE")
	{
		decl asdlg, asoptions, avalues, i;
		if ("OxPackDialog"("Store in Database",
				{	{ "Store in database" },
					{ "Residuals", CTL_CHECK, 0, "residuals" },
					{ "Forecasts", CTL_CHECK, rows(m_mForc) > 0 ? 0 : -1, "forecasts" },
					{ "Forecasts standard errors", CTL_CHECK, rows(m_mForc) > 0 ? 0 : -1, "forcse" },
					{ "Level forecasts", CTL_CHECK, rows(m_mForcLevel) > 0 ? 0 : -1, "forclevel" },
					{ "Level forecasts standard errors", CTL_CHECK, rows(m_mForcLevel) > 0 ? 0 : -1, "forclevelse" }
				},
				&asoptions, &avalues)
			)
		{
			for (i = 0; i < sizeof(asoptions); ++i)
				if (avalues[i] >= 1)
					DoStore(asoptions[i], TRUE);
		}
	}
	else
	{	// allow base class to process unhandled cases
		Modelbase::ReceiveDialog(sDialog, asOptions, aValues);
	}
}
Arfima::GetModelSettings()
{
	decl fixar, fixma, fixd, imean, dmean;
	[fixar, fixma, fixd, imean, dmean] = GetSettingValues();
	
	return
	{	{	m_cAR,			"p"		},
		{	m_cMA,			"q"		}, 	
		{	fixar,			"fixp"  },
		{	fixma,			"fixq"	},
		{	m_fDfixed~fixd,	"estd"	},
		{	imean~dmean,	"mean"	}
	};
}
Arfima::SetModelSettings(const aValues)
{
	decl i, car = 0, cma = 0;

	if (!isarray(aValues))
		return;
	for (i = 0; i < sizeof(aValues); ++i)
	{
		if (aValues[i][1] == "p")
			car = aValues[i][0], ARMA(car, cma);
		else if (aValues[i][1] == "q")
			cma = aValues[i][0], ARMA(car, cma);
		else if (aValues[i][1] == "fixp")
			FixAR(aValues[i][0]);
		else if (aValues[i][1] == "fixq")
			FixMA(aValues[i][0]);
		else if (aValues[i][1] == "estd")
		{	if (aValues[i][0][0] == 0)
				FreeD();
			else
				FixD(aValues[i][0][1]);
		}
		else if (aValues[i][1] == "mean")
		{
			if (aValues[i][0][0] == 2)
				FixMean(aValues[i][0][1]);
			else if (aValues[i][0][0] == 1)
				UseSampleMean();
			else
				FixMean(0);
		}
	}
}
